import torch
import jax
import jax.numpy as jnp
import numpy as np
from tqdm import tqdm
from .dataset import nTokenTypes, prepare_additional_molecule_feats, aa2tok_d, prepare_msa_masks

def get_msa_categorical_jacobian(
    model: torch.nn.Module,
    batch_input: torch.tensor, # One-hot encoded MSA size (S, L, 28 alphabet size)
    mask: torch.tensor,
    msa_mask: torch.tensor,
    model_type: str,
    k: int = 0, # Row index of sequence to compute jacobian over,
    mask_only: bool = False,
    eval_mode: bool = False,
    seq_weights: torch.tensor = None
):
    """
    Computes categorical Jacobian for AF3 MSA module or MSATransformer
    """
    assert model_type in ['AF3', 'MSATransformer']
    if eval_mode:
        model = model.eval()

    if model_type == 'AF3':
        # Add batch dimension to one-hot encoding
        batch_input = batch_input.unsqueeze(0) # size (1, S, L, A)

        # Get other inputs
        additional_molecule_feats = prepare_additional_molecule_feats(batch_input)

        # Query sequence length
        seq_length = batch_input.shape[2]
    else:
        seq_length = batch_input.shape[2] - 1

    # Load to GPU if available
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    mask = mask.to(device)
    msa_mask = msa_mask.to(device)
    batch_input = batch_input.to(device)
    if model_type == 'AF3':
        additional_molecule_feats = additional_molecule_feats.to(device)
    model = model.to(device)
    if seq_weights is not None:
        seq_weights = seq_weights.to(device)

    with torch.no_grad():
        # Function to get logits for the first sequence (only the 20 amino acids)
        # MSATransformer (ESM) uses 4-23 as AA tokens
        if model_type == 'AF3':
            aa_idx_offset = 0
            f = lambda x: model(msa = x, mask=mask, msa_mask=msa_mask, 
                                additional_molecule_feats=additional_molecule_feats,
                                seq_weights=seq_weights
                                )["logits"][0, k, :seq_length, :20].cpu().numpy()
        else:
            aa_idx_offset = 4
            f = lambda x: model(x)["logits"][0, k, 1:(seq_length+1), aa_idx_offset:20+aa_idx_offset].cpu().numpy()

        # Get logits for unperturbed MSA
        fx = f(batch_input) # L, 20
        # Initialize Jacobian matrix
        if mask_only:
            fx_h = np.zeros((seq_length, 1, seq_length, 20))
        else:
            fx_h = np.zeros((seq_length, 20, seq_length, 20))
        # For each residue
        res_offset = 0 if model_type == 'AF3' else 1 # ESM has a BOS token at the beginning
        for n in tqdm(range(seq_length)):
            # For all 20 mutations
            if mask_only:
                msa_h = torch.clone(batch_input)
                msa_h[0, k, n] = aa2tok_d['MASK']
                fx_h[n, 0] = f(msa_h)
            else:
                for a in range(20):
                    msa_h = torch.clone(batch_input)
                    msa_h[0, k, n+res_offset] = a + aa_idx_offset
                    fx_h[n, a] = f(msa_h)
        return fx - fx_h

def get_contacts_from_coev(x, symm=True, center=True, rm=1, mask_only=False):
    # convert jacobian (L,A,L,A) to contact map (L,L)
    j = x.copy()
    if center:
        if mask_only:
            for i in [l for l in range(4) if l != 1]: j -= j.mean(i,keepdims=True)
        else:
            for i in range(4): j -= j.mean(i,keepdims=True)
    j_fn = np.sqrt(np.square(j).sum((1,3)))
    np.fill_diagonal(j_fn,0)
    j_fn_corrected = do_apc(j_fn, rm=rm)
    if symm:
        j_fn_corrected = (j_fn_corrected + j_fn_corrected.T)/2
    return j_fn_corrected

def do_apc(x, rm=1):
    '''given matrix do apc correction'''
    # trying to remove different number of components
    # rm=0 remove none
    # rm=1 apc
    x = np.copy(x)
    if rm == 0:
        return x
    elif rm == 1:
        a1 = x.sum(0,keepdims=True)
        a2 = x.sum(1,keepdims=True)
        y = x - (a1*a2)/x.sum()
    else:
        # decompose matrix, rm largest(s) eigenvectors
        u,s,v = np.linalg.svd(x)
        y = s[rm:] * u[:,rm:] @ v[rm:,:]
    np.fill_diagonal(y,0)
    return y

def get_contacts_from_pairwise_rep(pairwise_rep: torch.tensor):
    """
    Computes L2-norm of (L, L, d) pairwise representation
    """
    if type(pairwise_rep) == np.ndarray:
        return np.linalg.norm(pairwise_rep, axis=2)
    else:
        return torch.norm(pairwise_rep, p=2, dim=2)

def get_coevolution(X, num_aa_types=23):
    '''given one-hot encoded MSA, return contacts'''
    Y = jax.nn.one_hot(X, num_aa_types)
    N,L,A = Y.shape
    Y_flat = Y.reshape(N,-1)
    # covariance
    c = jnp.cov(Y_flat.T)

    # inverse covariance
    shrink = 4.5/jnp.sqrt(N) * jnp.eye(c.shape[0])
    ic = jnp.linalg.inv(c + shrink)

    # partial correlation coefficient
    ic_diag = jnp.diag(ic)
    pcc = ic / jnp.sqrt(ic_diag[:,None] * ic_diag[None,:])
    
    raw = jnp.sqrt(jnp.square(pcc.reshape(L,A,L,A)[:,:20,:,:20]).sum((1,3)))
    i = jnp.arange(L)
    raw = raw.at[i,i].set(0)
    # do apc
    ap = (raw.sum(0, keepdims=True) * raw.sum(1, keepdims=True)) / raw.sum()
    return (raw - ap).at[i,i].set(0)
    






import re
import os
import numpy as np
import tempfile
import subprocess
import pickle
import einx

from glob import glob
from typing import List, Any, Union, Tuple
from pathlib import Path

from Bio import SeqIO
from Bio.Seq import Seq

from scipy.spatial.distance import cdist

import torch
from torch.utils.data import Dataset
from torch.nn.functional import one_hot

# Amino acid code to character
code2aa_d = {
    "ALA": "A",
    "ARG": "R",
    "ASN": "N",
    "ASP": "D",
    "CYS": "C",
    "GLU": "E",
    "GLN": "Q",
    "GLY": "G",
    "HIS": "H",
    "ILE": "I",
    "LEU": "L",
    "LYS": "K",
    "MET": "M",
    "PHE": "F",
    "PRO": "P",
    "SER": "S",
    "THR": "T",
    "TRP": "W",
    "TYR": "Y",
    "VAL": "V"
}

# Tokenize amino acids
aa2tok_d = {
    "A": 0, # ALA
    "R": 1, # ARG
    "N": 2, # ASN
    "D": 3, # ASP
    "C": 4, # CYS
    "E": 5, # GLU
    "Q": 6, # GLN
    "G": 7, # GLY
    "H": 8, # HIS
    "I": 9, # ILE
    "L": 10, # LEU
    "K": 11, # LYS
    "M": 12, # MET
    "F": 13, # PHE
    "P": 14, # PRO
    "S": 15, # SER
    "T": 16, # THR
    "W": 17, # TRP
    "Y": 18, # TYR
    "V": 19, # VAL
    "X": 20, # UNK
    "B": 21, # ASP or ASN
    "Z": 22, # GLU or GLN
    "U": 23, # SEC
    "O": 24, # PYL
    "-": 25, # GAP
    "<pad>": 26, # Padded positions
    "<mask>": 27, # Mask
}
tok2aa_d = {aa2tok_d[k]:k for k in aa2tok_d}
nTokenTypes = len(np.unique(list(aa2tok_d.values())))

# ESM token to amino acid
ESM_SEQUENCE_VOCAB = [
    "<cls>", "<pad>", "<eos>", "<unk>",
    "L", "A", "G", "V", "S", "E", "R", "T", "I", "D", "P", "K",
    "Q", "N", "F", "Y", "M", "H", "W", "C", "X", "B", "U", "Z",
    "O", ".", "-", "|",
    "<mask>"
]
esm_tok2aa_d = {ind: alph for ind, alph in enumerate(ESM_SEQUENCE_VOCAB)}
esm_aa2tok_d = {alph: ind for ind, alph in enumerate(ESM_SEQUENCE_VOCAB)}
esmtok_to_pairformertok_d = {esm_aa2tok_d[aa]: aa2tok_d[aa] if aa in aa2tok_d else -1 for aa in ESM_SEQUENCE_VOCAB}
esmtok_to_pairformertok_d[esm_aa2tok_d['<unk>']] = aa2tok_d['X'] # Handle unknown tokens with X
# MSA Pairformer was trained without insertions ".", so they are not part of the vocabulary
# Full list of tokens that do not appear in the MSA Pairformer are <cls>, <eos>, <unk>, <null_1>, <.>
# All are mapped to -1 using esmtok_to_pairformertok_d

def convert_tokens_esm2pairformer(batch_tokens: torch.Tensor, device: torch.device = torch.device('cpu')):
    # MSA Transformer uses a CLS token, but MSA Pairformer does not
    # We'll remove the first token to account for this disparity
    return torch.from_numpy(np.vectorize(esmtok_to_pairformertok_d.get)(batch_tokens.cpu().numpy()))[:, :, 1:].to(device)


class MSA:
    def __init__(
        self,
        msa_file_path: Union[str, Path], # Path to MSA file
        max_length: int = 1024, # Maximum length of the MSA (default is 1024)
        max_tokens: int = 1048576, # Maximum number of tokens in the MSA (default is 1024 * 1024)
        max_seqs: int = 1024, # Maximum number of sequences in the MSA (default is 1024)
        diverse_select_method = "hhfilter", # Method to select diverse sequences (default is "hhfilter")
        secondary_filter_method: str = "greedy",  # Options: "greedy" or "random" (default is "greedy")
        random_query: bool = False, # Whether to select a random query sequence (default is to use the first sequence as the query)
        min_query_coverage: float = 0.8, # Minimum query coverage if selecting a random query sequence
        parser_kwargs: dict = {
            "keep_insertions": False,
            "to_upper": False,
            "remove_lowercase_cols": False
        },
        hhfilter_kwargs: dict = {}
    ):
        self.msa_file_path = msa_file_path
        self.max_length = max_length
        self.max_tokens = max_tokens
        self.max_seqs = max_seqs
        self.diverse_select_method = diverse_select_method
        self.secondary_filter_method = secondary_filter_method
        self.random_query = random_query

        assert self.diverse_select_method in ["greedy", "hhfilter", "none"], "Diverse select method must be either 'greedy' or 'hhfilter', or 'none' if no diverse selection is desired"
        assert self.secondary_filter_method in ["greedy", "random", "none"], "Secondary filter method must be either 'greedy' or 'random', or 'none' if no secondary filter is desired"

        # Parse MSA file
        self.seq_l, self.ids_l = self.parse_a3m_file(**parser_kwargs)
        self.seq_a = self.seq_list_to_arr(self.seq_l)
        if self.random_query:
            # Select random sequence index and swap with query sequence
            # Ensure that randomly selected sequence has at least min_query_coverage (e.g. 80%) coverage
            valid_indices = np.where((self.seq_a != '-').sum(axis=-1) >= (self.seq_a.shape[-1] * min_query_coverage))[0]
            if len(valid_indices) > 0:
                random_idx = np.random.choice(valid_indices)
                self.seq_a[[0, random_idx]] = self.seq_a[[random_idx, 0]]
                # Swap ids
                tmp_id = self.ids_l[0]
                self.ids_l[0] = self.ids_l[random_idx]
                self.ids_l[random_idx] = tmp_id

        # Get random crop based on max length
        self.random_crop, self.random_crop_min, self.random_crop_max = self.get_random_crop(self.seq_a, max_length=self.max_length)
        # Select sequences based on max depth (greedy selection or hhfilter)
        # Determine maximum depth based on sequence length and max tokens
        # Currently capping the MSA to some number (max_seqs) of sequences
        # Could instead use min(full_msa_depth, max_tokens // sequence_length)
        sequence_length = self.random_crop.shape[1]
        self.subset_depth = min(self.max_tokens // sequence_length, self.max_seqs)
        if self.diverse_select_method == "none":
            self.select_diverse_msa = self.random_crop
            self.select_diverse_indices = np.arange(self.random_crop.shape[0])
        else:
            self.select_diverse_msa, self.select_diverse_indices = self.select_diverse(
                msa_a=self.random_crop,
                num_seqs=self.max_seqs,
                method=diverse_select_method,
                hhfilter_kwargs=hhfilter_kwargs,
                secondary_filter_method=self.secondary_filter_method
            )
        # Tokenize diverse MSA
        self.diverse_tokenized_msa = self.tokenize_msa(self.select_diverse_msa)
        self.n_diverse_seqs = self.diverse_tokenized_msa.shape[0]

    def compute_pssm(self, msa_a):
        # Get counts excluding padding token (26) and create probability distribution
        nSeqs, nPos = msa_a.shape
        probs = torch.zeros(nPos, aa2tok_d['<pad>'])  # Initialize output tensor [nPos, 26]

        # For each position in sequence
        for i in range(nPos):
            pos_tensor = msa_a[:, i]  # Get all sequences at this position
            mask = pos_tensor != aa2tok_d['<pad>']  # Mask padding tokens
            counts = torch.bincount(pos_tensor[mask], minlength=aa2tok_d['<pad>'])  # Count tokens at this position
            probs[i] = counts.float() / counts.sum()  # Convert to probabilities
        return probs

    def generate_random_sequences(self, probs, nSeqs):
        # Sample k sequences from probability distribution
        sampled_seqs = torch.zeros(nSeqs, probs.shape[0], dtype=torch.long)
        for i in range(probs.shape[0]):
            sampled_seqs[:, i] = torch.multinomial(probs[i], nSeqs, replacement=True)
        return sampled_seqs

    def parse_a3m_file(
        self,
        keep_insertions: bool = False,
        to_upper: bool = False,
        remove_lowercase_cols: bool = False, # Any lowercase columns in the query sequence
        **kwargs
    ):
        """
        Parse sequences from a3m file. 
        Returns list of full length sequences aligned to query sequence (top row)
        keep_insertions determines whether to keep insertions in sequences
        to_upper determines whether to convert sequences to uppercase (unnecessary if removing insertions)
        """
        seq_l = []
        ids_l = []
        valid_indices = None
        for record in SeqIO.parse(self.msa_file_path, "fasta"):
            sequence = str(record.seq)
            if remove_lowercase_cols:
                if valid_indices is None:
                    valid_indices = [i for i, aa in enumerate(sequence) if aa.isupper()]
                sequence = "".join([sequence[i] for i in valid_indices])
            if not keep_insertions:
                sequence = re.sub(r"[a-z]|\.|\*", "", sequence)
            if to_upper:
                sequence = sequence.upper()
            seq_l.append(sequence)
            ids_l.append(record.name)
        return seq_l, ids_l

    @property
    def inverse_covariance(self):
        if not hasattr(self, "_inverse_covariance"):
            # One hot encode
            msa_onehot = one_hot(self.diverse_tokenized_msa, num_classes=nTokenTypes)
            # Get shape
            n, l, a = msa_onehot.shape
            flat_msa = msa_onehot.reshape(n, -1)
            # Compute covariance
            c = torch.cov(flat_msa.T)
            # Inverse covariance
            shrink = 4.5 / np.sqrt(n) * torch.eye(c.shape[0])
            ic = torch.linalg.inv(c + shrink)
            # Sum across amino acid inverse covariances (1-21 in our case)
            ic = ic.reshape(l, a, l, a)[:, 1:21, :, 1:21].sum((1, 3))
            self._inverse_covariance = ic
        return self._inverse_covariance

    def tokenize_msa(self, msa_a):
        return torch.from_numpy(np.vectorize(aa2tok_d.get)(msa_a))

    def seq_list_to_arr(self, seq_l):
        return np.array([list(seq) for seq in seq_l])

    def get_random_crop(self, msa_a, max_length: int = 1024):
        seq_len = msa_a.shape[1]
        if seq_len <= max_length:
            return msa_a, 0, seq_len-1
        start = np.random.randint(0, seq_len - max_length)
        return self.seq_a[:, start:start+max_length], start, start+max_length-1

    def select_diverse(self, msa_a, num_seqs: int, method: str = "hhfilter", hhfilter_kwargs: dict = {}, secondary_filter_method: str = "greedy"):
        assert method in ['greedy', 'hhfilter'], "Method must be either 'greedy' or 'hhfilter'"
        if method == 'greedy':
            return self.greedy_select(msa_a, num_seqs)
        elif method == 'hhfilter':
            hhfilter_kwargs = {**hhfilter_kwargs, "diff": num_seqs}
            filtered_msa, kept_indices = self.hhfilter_select(msa_a, **hhfilter_kwargs)
            # If hhfilter returns more sequences than maximum depth,
            # maximize diversity with maximum depth
            if num_seqs < filtered_msa.shape[0]:
                if secondary_filter_method == "greedy":
                    # Greedily select for maximum MSA diversity
                    filtered_msa, greedy_kept_indices = self.greedy_select(filtered_msa, num_seqs)
                    kept_indices = [kept_indices[i] for i in greedy_kept_indices]
                elif secondary_filter_method == "random":
                    # Randomly select num_seqs sequences (always include query sequence)
                    random_indices = np.concatenate([[0], np.random.choice(np.arange(1, filtered_msa.shape[0]), num_seqs-1, replace=False)])
                    filtered_msa = filtered_msa[random_indices]
                    kept_indices = [kept_indices[i] for i in random_indices]
                elif secondary_filter_method == "none":
                    pass
                else:
                    raise ValueError(f"Secondary filter method must be either 'greedy' or 'random', got {secondary_filter_method}")
            return filtered_msa, kept_indices
        else:
            raise ValueError("Method must be either 'greedy' or 'hhfilter'")

    def greedy_select(self, msa_a, num_seqs: int):
        tokenized_msa = self.tokenize_msa(msa_a)
        # Already below depth threshold
        curr_depth = msa_a.shape[0]
        if curr_depth <= num_seqs:
            return msa_a, np.arange(curr_depth)
        # Greedily maximize diversity
        all_indices = np.arange(curr_depth)
        indices = [0]
        pairwise_distances = np.zeros((0, curr_depth))
        for _ in range(num_seqs - 1):
            dist = cdist(tokenized_msa[indices[-1:]], tokenized_msa, "hamming")
            pairwise_distances = np.concatenate([pairwise_distances, dist])
            shifted_distance = np.delete(pairwise_distances, indices, axis=1).mean(0)
            shifted_index = np.argmax(shifted_distance)
            index = np.delete(all_indices, indices)[shifted_index]
            indices.append(index)
        indices = sorted(indices)
        return msa_a[indices], indices

    def hhfilter_select(
        self,
        msa_a,
        M = "a3m",
        seq_id: int=90,
        diff: int=0, # Number of sequences
        cov: int=70,
        qid: int=15,
        qsc: float=-20.0,
        maxseq: int=None,
        binary="hhfilter",
    ):
        # Get tmp directory from environment
        tmpdir_base = os.environ.get("TMPDIR", "/tmp/")
        with tempfile.TemporaryDirectory(dir=tmpdir_base) as tmpdirname:
            tmpdir = Path(tmpdirname)
            random_prefix = ''.join(np.random.choice(list('abcdefghijklmnopqrstuvwxyz0123456789'), size=8))
            fasta_file = tmpdir / f"{random_prefix}.input.fasta"
            fasta_file.write_text(
                "\n".join([f">{i}\n{''.join(seq)}" for i, seq in enumerate(msa_a)])
            )
            output_file = tmpdir / f"{random_prefix}.output.fasta"
            command = " ".join(
                [
                    f"{binary}",
                    f"-i {fasta_file}",
                    f"-M {M}",
                    f"-o {output_file}",
                    f"-id {seq_id}",
                    f"-diff {diff}",
                    f"-cov {cov}",
                    f"-qid {qid}",
                    f"-qsc {qsc}",
                ]
            ).split(" ")
            if maxseq is not None:
                command.append(f"-maxseq {maxseq}")
            result = subprocess.run(command, capture_output=True)
            result.check_returncode()
            with output_file.open() as f:
                indices = [int(line[1:].strip()) for line in f if line.startswith(">")]
            return msa_a[indices], indices
            

class MSADataset(Dataset):
    def __init__(
        self,
        msa_dir=None,
        msa_paths = None,
        max_seq_length = 1024,
        max_msa_depth = 1024,
        max_tokens = 2**17,
        min_depth = 4,
        transform = None,
        random_query: bool = False,
        min_query_coverage: float = 0.8
    ):
        """
        data_dir is the parent directory that stores all of the MSA .a3m files
        """
        assert (msa_dir is not None) or (msa_paths is not None), "Must provide MSA paths or directory path"
        super().__init__()
        self.msa_dir = msa_dir
        self.transform = transform
        self.max_seq_length = max_seq_length
        self.max_msa_depth = max_msa_depth
        self.max_tokens = max_tokens
        self.min_depth = min_depth
        self.random_query = random_query
        self.min_query_coverage = min_query_coverage
        if msa_paths is None:
            self.msa_paths = glob(os.path.join(msa_dir, "*/a3m/*.a3m"))
        else:
            self.msa_paths = msa_paths
        
    def __len__(self):
        return len(self.msa_paths)

    def __getitem__(self, idx):
        res_d = {}
        # Get MSA file path
        msa_path = self.msa_paths[idx]
        # Create MSA object
        msa = MSA(
            msa_path,
            max_length = self.max_seq_length,
            max_seqs = self.max_msa_depth,
            max_tokens = self.max_tokens,
            diverse_select_method = "hhfilter",
            random_query = self.random_query,
            min_query_coverage = self.min_query_coverage    
        )
        # Get tokenized MSA
        res_d['tokenized_msa'] = msa.diverse_tokenized_msa
        # Add MSA depth
        res_d['n_diverse_seqs'] = msa.n_diverse_seqs
        # Add file path
        res_d['file_path'] = msa_path
        return res_d     

def msa_mlm(
    msa_t: torch.tensor, 
    mask_tok: int = aa2tok_d['<mask>'],
    pad_tok: int = aa2tok_d['<pad>'],
    mask_prob: float = 0.15,
    mask_ratio: float = 0.8,
    mutate_ratio: float = 0.1,
    keep_ratio: float = 0.1,
    mutate_pssm: bool = True,
    mutate_tok_low: int = 0,
    mutate_tok_high: int = 19,
    query_only: bool = False
):
    # Create masked input (don't mask padding tokens)
    masked_msas = msa_t.clone()
    flat_msas = masked_msas.view(-1)
    if query_only:
        nMSAs, depth, length = masked_msas.shape
        intervals_t = torch.arange(nMSAs).reshape(nMSAs, 1) * depth * length
        non_pad_indices = (torch.arange(length).unsqueeze(0).repeat(nMSAs, 1) + intervals_t).flatten()
        non_pad_indices = non_pad_indices[flat_msas[non_pad_indices] != pad_tok]
    else:
        non_pad_indices = torch.nonzero(flat_msas != pad_tok, as_tuple=False).view(-1)
    
    # Calculate the number of positions to mask
    num_mask = int(len(non_pad_indices) * mask_prob)

    # Generate mask indices
    mlm_indices = np.random.choice(non_pad_indices.numpy(), num_mask, replace=False)
    mask_ub = int(mask_ratio * num_mask)
    mask_indices = mlm_indices[:mask_ub]
    mutate_ub = mask_ub + int(mutate_ratio * num_mask)
    mutate_indices = mlm_indices[mask_ub : mutate_ub]
    keep_indices = mlm_indices[mutate_ub:]

    # Apply MLM (masking and mutating)
    masked_msas.view(-1)[mask_indices] = torch.tensor(mask_tok)
    if mutate_pssm:
        # Get column indices of mutate_indices
        batch_indices, seq_indices, pos_indices = torch.unravel_index(torch.tensor(mutate_indices), msa_t.shape)
        # Compute PSSM of mutate_indices (exclude padding tokens)
        counts_t = torch.nn.functional.one_hot(msa_t, num_classes=msa_t.shape[-1]).sum(dim=1)[:, :, :26]
        pssm = counts_t / counts_t.sum(dim=-1, keepdim=True)
        probs_t = pssm[batch_indices, pos_indices]
        probs_t = probs_t[mutate_indices]
        new_toks_t = torch.multinomial(probs_t, num_samples=1)
        # new_toks_t = torch.stack([torch.multinomial(probs_t[i], num_samples=1) for i in range(len(mutate_indices))]).squeeze()
        # Apply mutations
        masked_msas.view(-1)[mutate_indices] = new_toks_t
    else:
        masked_msas.view(-1)[mutate_indices] = torch.randint_like(input = masked_msas.view(-1)[mutate_indices], low=mutate_tok_low, high=mutate_tok_high+1)

    # Return masked MSA and indices of tokens to predict
    return masked_msas, mlm_indices

class CollateAFBatch():
    def __init__(
        self,
        max_seq_length,
        max_seq_depth,
        min_seq_depth,
        pad_tok=aa2tok_d['<pad>'],
        mask_tok=aa2tok_d['<mask>'], 
        mask_prob=0.15,
        mask_ratio=0.8,
        mutate_ratio=0.1,
        keep_ratio=0.1,
        tok_low=0,
        tok_high=25,
        query_only=False,
        mutate_pssm=False,
    ):
        self.max_seq_depth = max_seq_depth
        self.max_seq_length = max_seq_length
        self.min_seq_depth = min_seq_depth
        self.pad_tok = pad_tok
        self.mask_tok = mask_tok
        self.mask_prob = mask_prob
        self.mask_ratio = mask_ratio
        self.mutate_ratio = mutate_ratio
        self.keep_ratio = keep_ratio
        self.tok_low = tok_low
        self.tok_high = tok_high
        self.mutate_pssm = mutate_pssm
        self.query_only = query_only

    def __call__(self, batch):
        # Initialize output dictionary
        output_dict = {}

        # Skip MSAs with too few sequences
        valid_msa = [b['n_diverse_seqs'] >= self.min_seq_depth for b in batch]
        if not any(valid_msa):
            return None
        # Get batch size
        og_batch_size = len(batch)
        effective_batch_size = sum(valid_msa)
        
        # Match sequence lengths and stack tensors
        tokenized_msa_l = [batch[i]['tokenized_msa'] for i in range(og_batch_size) if valid_msa[i]]
        # shape = [batch_size] + np.max([seq.shape for seq in tokenized_msa_l], 0).tolist()
        shape = [effective_batch_size, self.max_seq_depth, self.max_seq_length]
        dtype = tokenized_msa_l[0].dtype
        if isinstance(tokenized_msa_l[0], np.ndarray):
            msas = np.full(shape, self.pad_tok, dtype=dtype)
        elif isinstance(tokenized_msa_l[0], torch.Tensor):
            msas = torch.full(shape, self.pad_tok, dtype=dtype)
        for msa, seq in zip(msas, tokenized_msa_l):
            msaslice = tuple(slice(dim) for dim in seq.shape)
            msa[msaslice] = seq
        output_dict['msas'] = msas
    
        # Create masked input
        masked_msas, mlm_indices = msa_mlm(msas, mask_tok=self.mask_tok, pad_tok=self.pad_tok, mask_prob=self.mask_prob, mask_ratio=self.mask_ratio, 
                                           mutate_ratio=self.mutate_ratio, keep_ratio=self.keep_ratio, mutate_tok_low=self.tok_low, 
                                           mutate_tok_high=self.tok_high, query_only=self.query_only, mutate_pssm=self.mutate_pssm)
        # One-hot encode masked/mutated MSA
        masked_msas_onehot = one_hot(masked_msas, num_classes = nTokenTypes)
        output_dict['msas_onehot'] = masked_msas_onehot
        if self.query_only:
            _, _, n_pos = msas.shape
            batch_idx, seq_idx, pos_idx = np.unravel_index(mlm_indices, msas.shape)
            query_only_pos_idx = pos_idx + batch_idx * n_pos
            output_dict['masked_idx'] = query_only_pos_idx
        else:
            output_dict['masked_idx'] = mlm_indices
        output_dict['unmasked_msas_onehot'] = one_hot(msas, num_classes = nTokenTypes)
    
        # Initialize additional molecule features
        molecule_feats = prep_molecule_feats(output_dict['msas_onehot'])
        output_dict['molecule_feats'] = molecule_feats
        # Store file path
        output_dict['file_path'] = [batch[i]['file_path'] for i in range(og_batch_size) if valid_msa[i]]
        # Store MSA depth
        output_dict['msa_depths'] = torch.tensor([batch[i]['n_diverse_seqs'] for i in range(og_batch_size) if valid_msa[i]])
        # Prepare masks
        mask, msa_mask, full_mask, pairwise_mask = prepare_msa_masks(output_dict['msas'])
        output_dict['mask'] = mask
        output_dict['msa_mask'] = msa_mask
        output_dict['full_mask'] = full_mask
        output_dict['pairwise_mask'] = pairwise_mask
        return output_dict

def prep_molecule_feats(msa_input):
    batch_size = msa_input.shape[0]
    seq_len = msa_input.shape[2]
    molecule_feats = torch.stack([
        torch.ones(size=(batch_size, seq_len)), # molecule_idx
        torch.arange(seq_len).unsqueeze(0).repeat(batch_size, 1), # token_index
        torch.zeros(size=(batch_size, seq_len)), # molecule_idx
        torch.zeros(size=(batch_size, seq_len)),# entity_id,
        torch.zeros(size=(batch_size, seq_len)), # sym_id
    ], dim=-1).reshape((batch_size, seq_len, 5))
    return molecule_feats

def get_relative_positions(msa_input):
    batch_size = msa_input.shape[0]
    seq_len = msa_input.shape[2]
    token_indices = torch.arange(seq_len).unsqueeze(0).repeat(batch_size, 1)
    return token_indices

def onehot_msa(msa_input, device=torch.device('cpu')):
    return torch.nn.functional.one_hot(msa_input, num_classes=len(aa2tok_d)).to(device)

def prepare_msa_masks(msa_input, device=torch.device('cpu')):
    # msa_input is of shape [b, s, n]
    mask = (msa_input != aa2tok_d['<pad>']).any(dim=1) # [b, n]
    msa_mask = (msa_input != aa2tok_d['<pad>']).any(dim=2) # [b, s]
    full_mask = (msa_input != aa2tok_d['<pad>']) # [b, s, n]
    pairwise_mask = einx.logical_and("... i, ... j -> ... i j", mask, mask) # [b, n, n]
    return mask.to(device), msa_mask.to(device), full_mask.to(device), pairwise_mask.to(device)

def prepare_inputs(batch, device):
    msa_repr = batch['msas_onehot'].float().to(device)
    mask, msa_mask, full_mask, pairwise_mask = prepare_msa_masks(batch['msas'])
    msa_mask = msa_mask.to(device)
    mask = mask.to(device)
    full_mask = full_mask.to(device)
    pairwise_mask = pairwise_mask.to(device)
    molecule_feats = batch['molecule_feats'].to(device)
    return msa_repr, mask, msa_mask, full_mask, pairwise_mask, molecule_feats

def prepare_inputs_bf16(batch, device):
    msa_repr = batch['msas_onehot'].bfloat16().to(device)
    mask, msa_mask, full_mask, pairwise_mask = prepare_msa_masks(batch['msas'])
    msa_mask = msa_mask.to(device)
    mask = mask.to(device)
    full_mask = full_mask.to(device)
    pairwise_mask = pairwise_mask.to(device)
    molecule_feats = batch['molecule_feats'].bfloat16().to(device)
    return msa_repr, mask, msa_mask, full_mask, pairwise_mask, molecule_feats

def create_msa_subset_mask(subset_msa_idx_l, L):
    b = len(subset_msa_idx_l)
    mask = torch.zeros(b, L, L)
    for i in range(b):
        indices = torch.tensor(subset_msa_idx_l[i])
        row_mask = torch.zeros(L, dtype=torch.bool)
        row_mask[indices] = True
        mask[i] = row_mask.unsqueeze(0) & row_mask.unsqueeze(1)
    return mask > 0


class trRosettaContactMSADataset(Dataset):
    def __init__(
        self,
        paired_paths_l: List[Tuple[Union[str, Path], Union[str, Path]]],
        max_seq_length: int = 1024,
        max_msa_depth: int = 1024,
        min_msa_depth: int = 8,
        max_tokens: int = 2**17,
        random_query: bool = False,
        min_query_coverage: float = 0.9
    ):
        self.paired_paths_l = paired_paths_l
        self.max_seq_length = max_seq_length
        self.max_msa_depth = max_msa_depth
        self.max_tokens = max_tokens
        self.random_query = random_query
        self.min_query_coverage = min_query_coverage

    def __len__(self):
        return len(self.paired_paths_l)

    def __getitem__(self, idx):
        # Get MSA file path
        msa_path, npz_file_path = self.paired_paths_l[idx]
        # Create MSA mobject
        msa = MSA(
            msa_path,
            max_length = self.max_seq_length,
            max_seqs = self.max_msa_depth,
            max_tokens = self.max_tokens,
            diverse_select_method = "hhfilter",
            random_query = self.random_query,
            min_query_coverage = self.min_query_coverage
        )
        res_d = {}
        # Get tokenized MSA
        res_d['tokenized_msa'] = msa.diverse_tokenized_msa
        res_d['n_diverse_seqs'] = msa.n_diverse_seqs
        # Get contact map
        npz_obj = np.load(npz_file_path)
        res_d['contact_map'] = torch.tensor((npz_obj['dist6d'] > 0) & (npz_obj['dist6d'] < 8))
        # Add file path
        res_d['msa_file_path'] = msa_path
        res_d['npz_file_path'] = npz_file_path
        # Get random crop bounds
        res_d['msa_crop_bounds'] = (msa.random_crop_min, msa.random_crop_max)
        return res_d

class CollatetrRosettaContactMSABatch():
    def __init__(
        self, 
        max_seq_length: int, 
        max_seq_depth: int,
        min_seq_depth: int,
        pad_tok: int = aa2tok_d['<pad>']
    ):
        self.max_seq_depth = max_seq_depth
        self.max_seq_length = max_seq_length
        self.min_seq_depth = min_seq_depth
        self.pad_tok = pad_tok

    def __call__(self, batch):
        # Initialize output dictionary
        output_dict = {}

        # Skip MSAs with too few sequences
        valid_msa = [b['n_diverse_seqs'] >= self.min_seq_depth for b in batch]
        if not any(valid_msa):
            return None
    
        # Get batch size
        og_batch_size = len(batch)
        effective_batch_size = sum(valid_msa)
        
        # Match sequence lengths and stack tensors
        tokenized_msa_l = [batch[i]['tokenized_msa'] for i in range(og_batch_size) if valid_msa[i]]
        shape = [effective_batch_size, self.max_seq_depth, self.max_seq_length]
        dtype = tokenized_msa_l[0].dtype
        if isinstance(tokenized_msa_l[0], np.ndarray):
            msas = np.full(shape, self.pad_tok, dtype=dtype)
        elif isinstance(tokenized_msa_l[0], torch.Tensor):
            msas = torch.full(shape, self.pad_tok, dtype=dtype)
        for msa, seq in zip(msas, tokenized_msa_l):
            msaslice = tuple(slice(dim) for dim in seq.shape)
            msa[msaslice] = seq
        output_dict['msas'] = msas
        # Match lengths of contact maps
        contact_map_l = [batch[i]['contact_map'].float() for i in range(og_batch_size) if valid_msa[i]]
        msa_crop_bounds_l = [batch[i]['msa_crop_bounds'] for i in range(og_batch_size) if valid_msa[i]]
        contact_map_l = [contact_map_l[i][msa_crop_bounds_l[i][0]:msa_crop_bounds_l[i][1], msa_crop_bounds_l[i][0]:msa_crop_bounds_l[i][1]] for i in range(len(contact_map_l))]
        shape = [effective_batch_size, shape[-1], shape[-1]]
        dtype = contact_map_l[0].dtype
        if isinstance(contact_map_l[0], np.ndarray):
            contacts = np.full(shape, -1, dtype=dtype)
        elif isinstance(contact_map_l[0], torch.Tensor):
            contacts = torch.full(shape, -1, dtype=dtype)
        for i, contact_map in enumerate(contact_map_l):
            contacts[i, :contact_map.shape[0], :contact_map.shape[1]] = contact_map
        
        output_dict['contact_map'] = contacts

        # One-hot encode MSA
        msas_onehot = one_hot(msas, num_classes = nTokenTypes)
        output_dict['msas_onehot'] = msas_onehot
        # Initialize additional molecule features
        molecule_feats = prep_molecule_feats(output_dict['msas_onehot'])
        output_dict['molecule_feats'] = molecule_feats
        # Store file path
        output_dict['msa_file_path'] = [batch[i]['msa_file_path'] for i in range(og_batch_size) if valid_msa[i]]
        output_dict['npz_file_path'] = [batch[i]['npz_file_path'] for i in range(og_batch_size) if valid_msa[i]]
        # Prepare masks
        mask, msa_mask, full_mask, pairwise_mask = prepare_msa_masks(output_dict['msas'])
        output_dict['mask'] = mask
        output_dict['msa_mask'] = msa_mask
        output_dict['full_mask'] = full_mask
        output_dict['pairwise_mask'] = pairwise_mask
        return output_dict


from torch import Tensor
from jaxtyping import (
    Float as JaxFloat,
    Int as JaxInt,
    Bool as JaxBool,
    Shaped as JaxShaped,
    jaxtyped
)

def always(value):
    def inner(*args, **kwargs):
        return value
    return inner

def identity(t):
    return t

class TorchTyping:
    def __init__(self, abstract_dtype):
        self.abstract_dtype = abstract_dtype

    def __getitem__(self, shapes: str):
        return self.abstract_dtype[Tensor, shapes]

# PyTorch-compatible type annotations
Shaped = TorchTyping(JaxShaped)
Float = TorchTyping(JaxFloat)
Int = TorchTyping(JaxInt)
Bool = TorchTyping(JaxBool)

# Type checking is disabled by default
should_typecheck = False
typecheck = identity
beartype_isinstance = always(True)
__all__ = [
    'Float',
    'Int',
    'Bool',
    'typecheck',
    'beartype_isinstance'
]



#### Adapted from alphafold3-pytorch repository (https://github.com/lucidrains/alphafold3-pytorch/)
import torch
import torch.nn.functional as F
from torch import nn, sigmoid
from torch import Tensor
from torch.nn import Module, ModuleList, Linear, Sequential, ModuleDict, LayerNorm
from torch.utils.checkpoint import checkpoint

import einx
import re
import numpy as np
from math import pi, sqrt, exp
from functools import partial, wraps
from einops import rearrange, einsum, repeat, reduce, pack, unpack
from einops.layers.torch import Rearrange
from typing import List, Literal, Tuple, Optional

from .core import LinearNoBias, PreLayerNorm, Transition, exists, RMSNorm
from .outer_product import OuterProduct
from .regression import ContactHead, LMHead, LogisticRegressionContactHead, MultiLayerLogisticRegressionContactHead
from .pairwise_operations import MSAPairWeightedAveraging, PairwiseBlock
from .positional_encoding import RelativePositionEncoding
from .alphafold3_attention import Attention
from .alphafold3_custom_typing import (
    Float,
    Bool,
    typecheck
)

###############################################################
# Partially ablated MSA modules for MLM or inv cov. objective #
###############################################################
class CoreModule(Module):
    """
    Removes pair representation input and operations, but maintains an outer product mean
    tensor and updates in each layer using a residual connection.
    Also performs pair-weighted averaging + transition after
    pair representation updates (AF3 performs this before pair rep updates)
    """
    def __init__(
        self,
        *,
        depth: int = 4, # Number of blocks in stack
        dim_pairwise: int = 128,
        dim_msa: int = 64,
        opm_kwargs: dict = dict(
            dim_opm_hidden = 32,
            outer_product_flavor = "vanilla",
            seq_attn = False,
            dim_qk = None,
            chunk_size = None,
            return_seq_weights = False,
            return_attn_logits = False,
            lambda_init = None,
            eps = 1e-32,
            nSeqs = None,
            nResidues = None
        ),
        pwa_kwargs: dict = dict(
            heads = 8,
            dim_head = 32,
            dropout = 0.15,
            dropout_type = "row",
            return_attn_weights = False
        ),
        pairwise_block_kwargs: dict = dict(
            tri_mult_dim_hidden = None,
            tri_attn_dim_head = 32,
            tri_attn_heads = 4,
            dropout_row_prob = 0,
            dropout_col_prob = 0,
            use_triangle_updates = True,
            use_pair_updates = False
        ),
        dim_logits = 26,
        drop_last_msa_update = False,
        return_after_layer_idx = None,
        return_all_pairwise_repr = False
    ):
        super().__init__()

        # Store parameters
        self.dim_pairwise = dim_pairwise
        self.return_seq_weights = opm_kwargs['return_seq_weights'] if 'return_seq_weights' in opm_kwargs else False
        self.return_attn_weights = pwa_kwargs['return_attn_weights'] if 'return_attn_weights' in pwa_kwargs else False
        self.return_after_layer_idx = return_after_layer_idx
        self.return_all_pairwise_repr = return_all_pairwise_repr

        # Automatically assign lambda init if not provided
        if ('lambda_init' not in opm_kwargs) or (opm_kwargs['lambda_init'] is None):
            auto_lambda_init = True
        else:
            auto_lambda_init = False
            lambda_init = opm_kwargs['lambda_init']
        if 'lambda_init' in opm_kwargs:
            opm_kwargs.pop("lambda_init")

        # Initialize module stack
        layers = ModuleList([])
        for layer_idx in range(depth):
            curr_module_list = ModuleList()

            # MSA pair weighted averaging with gating -> transition
            msa_pair_weighted_avg = MSAPairWeightedAveraging(
                dim_msa = dim_msa, 
                dim_pairwise = dim_pairwise,
                **pwa_kwargs
            )
            msa_pre_ln = partial(PreLayerNorm, dim = dim_msa)
            curr_module_list.append(msa_pair_weighted_avg)
            curr_module_list.append(msa_pre_ln(Transition(dim = dim_msa)))

            # Outer product
            if auto_lambda_init and ('differential' in opm_kwargs['outer_product_flavor']):
                lambda_init = 0.8 - 0.6 * exp(-0.3 * layer_idx)
                lambda_init = torch.tensor(lambda_init, dtype=torch.bfloat16)
                # print(f"Layer {layer_idx} lambda init: {lambda_init}")
            else:
                lambda_init = None
            opm = OuterProduct(
                dim_msa = dim_msa,
                dim_pairwise = dim_pairwise,
                lambda_init = lambda_init,
                **opm_kwargs
            )
            curr_module_list.append(opm)

            # Pairwise representation update block with triangle attention
            pairwise_block = PairwiseBlock(
                dim_pairwise = dim_pairwise,
                **pairwise_block_kwargs
            )
            curr_module_list.append(pairwise_block)

            # Append all blocks of current layer to module list
            layers.append(curr_module_list)

        # Store layers in class object
        self.layers = layers

        # If we want to do a final MSA update
        self.final_msa_pwa = None
        self.final_msa_transition = None
        if not drop_last_msa_update:
            # MSA pair weighted averaging with gating
            self.final_msa_pwa = MSAPairWeightedAveraging(
                dim_msa = dim_msa, 
                dim_pairwise = dim_pairwise,
                **pwa_kwargs
            )
            # Transition module
            msa_pre_ln = partial(PreLayerNorm, dim = dim_msa)
            self.final_msa_transition = msa_pre_ln(Transition(dim = dim_msa))

        # Other parameters
        self.register_buffer('zero', torch.tensor(0.), persistent = False)

    def turn_off_seq_attn(self):
        for layer in self.layers:
            for module in layer:
                if isinstance(module, OuterProduct):
                    module.opm.seq_attn = False

    def turn_on_seq_attn(self):
        for layer in self.layers:
            for module in layer:
                if isinstance(module, OuterProduct):
                    module.opm.seq_attn = True
    
    def forward(
        self,
        msa: Float['b s n dm'],
        pairwise_repr: Float['b n n dp'],
        mask: Bool['b n'] | None = None, # Column mask
        msa_mask: Bool['b s'] | None = None, # Row mask,
        full_mask: Bool['b s n'] | None = None,
        pairwise_mask: Bool['b n n'] | None = None,
        seq_weights: Float['b s'] | None = None
    ) -> Float['b s n dm']:
        # Track seq weights
        seq_weights_list_d = {}
        attn_gates_list_d = {}
        attn_weights_list_d = {}
        pairwise_repr_l = []
        layer_idx = 0
        attn_weights = None
        # Pass MSA through each layer of the MSA module stack
        for (
            msa_pair_weighted_avg,
            msa_transition,
            outer_product,
            pairwise_block
        ) in self.layers:
            # Pair weighted averaging (with residual connection)
            msa_residual, attn_weights, attn_gates = msa_pair_weighted_avg(msa = msa, pairwise_repr = pairwise_repr, attn_weights = attn_weights, mask = mask)
            msa = msa + msa_residual
            if self.return_seq_weights and exists(attn_gates):
                attn_gates_list_d[f"layer_{layer_idx}"] = attn_gates
            if self.return_attn_weights and exists(attn_weights):
                attn_weights_list_d[f"layer_{layer_idx}"] = attn_weights
            del attn_gates, msa_residual
            msa = msa + msa_transition(msa)

            # Compute outer product mean (with residual connection)
            update_pairwise_repr, norm_weights, attn_weights = outer_product(msa, mask = mask, msa_mask = msa_mask, full_mask = full_mask, pairwise_mask = pairwise_mask, seq_weights = seq_weights)
            if self.return_seq_weights:
                seq_weights_list_d[f"layer_{layer_idx}"] = norm_weights
            pairwise_repr = pairwise_repr + update_pairwise_repr
            del update_pairwise_repr

            # Pairwise representation block (# Debug: make sure we haven't already normalized the pairwise repr)
            pairwise_repr = pairwise_block(pairwise_repr = pairwise_repr, mask = mask)
            layer_idx += 1
            if self.return_after_layer_idx is not None and layer_idx > self.return_after_layer_idx:
                break
            if self.return_all_pairwise_repr:
                pairwise_repr_l.append(pairwise_repr)

        if exists(self.final_msa_pwa) and (self.return_after_layer_idx is None):
            msa_residual, attn_weights, attn_gates = self.final_msa_pwa(msa = msa, pairwise_repr = pairwise_repr, mask = mask, attn_weights = attn_weights)
            msa = msa + msa_residual
            if self.return_seq_weights:
                attn_gates_list_d[f"layer_{layer_idx+1}"] = attn_gates
            del attn_gates, msa_residual
            msa = msa + self.final_msa_transition(msa)
        if not self.return_seq_weights:
            seq_weights_list_d, attn_weights_list_d = None, None
        if not self.return_attn_weights:
            attn_weights_list_d = None
        if self.return_all_pairwise_repr:
            pairwise_repr = torch.stack(pairwise_repr_l, dim=1)
        return msa, pairwise_repr, seq_weights_list_d, attn_gates_list_d, attn_weights_list_d

class MSAPairformer(Module):
    """
    3 main components:
        1) Core module
            a) Takes as input i) MSA representation; ii) Pair representation
            b) Iteratively updates MSA representation and pair representation
            c) Outputs refined MSA and pair representations
        2) MSA language model head
            b) Takes as input the final MSA representation
            c) Outputs the logits for the MSA language model
        3) Pairformer stack (optional)
    Trained using a variety of potential objectives:
        1) MSA language model (corruption via masking + mutation -> original MSA)
        2) MRF cross-entropy loss
        3) Alignment (Unalign MSA --> Realign MSA)
        4) Combination of multiple
    """
    def __init__(
        self,
        *,
        dim_msa_input: int = 28,
        dim_pairwise = 128,
        dim_msa = 64,
        dim_logits = 26,
        msa_module_kwargs: dict = dict(
            depth = 4, # Number of blocks in stack
            opm_kwargs = dict(
                dim_opm_hidden = 32,
                outer_product_flavor = "vanilla",
                seq_attn = False,
                dim_qk = None,
                chunk_size = None,
                return_seq_weights = False,
                lambda_init = None,
                eps = 1e-32
            ),
            pwa_kwargs = dict(
                heads = 8,
                dim_head = 32,
                dropout = 0.15,
                dropout_type = "row",
                return_attn_weights = False
            ),
            pairwise_block_kwargs = dict(
                tri_mult_dim_hidden = None,
                tri_attn_dim_head = 32,
                tri_attn_heads = 4,
                dropout_row_prob = 0,
                dropout_col_prob = 0,
                use_triangle_updates = True,
                use_pair_updates = False
            ),
            return_after_layer_idx = None,
            return_all_pairwise_repr = False
        ),
        basic_relative_position_encoding_kwargs: dict = dict(
            r_max = 32, # Maximum relative distance between two residues
        ),
        relative_position_encoding_kwargs: dict = dict(
            r_max = 32, # Maximum relative distance between two residues
            s_max = 2, # Maximum relative "chain" distance between two chains
        ),
        use_basic_relative_position_encoding: bool = False,
        return_msa_repr: bool = False,
        return_pairwise_repr: bool = False,
        query_only: bool = False
    ):
        super().__init__()
        self.dim_pairwise = dim_pairwise
        self.dim_msa = dim_msa
        # Initializing pairwise representation
        # Relative position encoding
        if use_basic_relative_position_encoding:
            self.relative_position_encoding = BasicRelativePositionEncoding(
                dim_out = dim_pairwise,
                **basic_relative_position_encoding_kwargs
            )
        else:
            self.relative_position_encoding = RelativePositionEncoding(
                dim_out = dim_pairwise,
                **relative_position_encoding_kwargs
            )
        # Token bonds for flanking residues (Algorithm 1 line 5)
        # z^{init}_{ij} += LinearNoBias(f^{token_bonds}_{ij})
        self.token_bond_to_pairwise_feat = Sequential(
            Rearrange('... -> ... 1'),
            LinearNoBias(1, dim_pairwise)
        )

        # Initial MSA projection
        self.msa_init_proj = LinearNoBias(dim_msa_input, dim_msa) if exists(dim_msa_input) else nn.Identity()
        
        # MSA module
        self.msa_module = CoreModule(
            dim_msa = dim_msa,
            dim_pairwise = dim_pairwise,
            **msa_module_kwargs
        )

        # Projection layer for logits
        self.lm_head = LMHead(
            dim_msa,
            dim_logits
        )
        # Initialize so that predictions are close to uniform across amino acids
        torch.nn.init.xavier_uniform_(self.lm_head.weight)
        torch.nn.init.constant_(self.lm_head.bias, 0)

        # Flag whether to return MSA and pairwise representations
        self.use_basic_relative_position_encoding = use_basic_relative_position_encoding
        self.return_msa_repr = return_msa_repr
        self.return_pairwise_repr = return_pairwise_repr
        self.return_after_layer_idx = msa_module_kwargs['return_after_layer_idx'] if 'return_after_layer_idx' in msa_module_kwargs else None
        self.query_only = query_only
        self.return_seq_weights = msa_module_kwargs['opm_kwargs']['return_seq_weights'] if 'return_seq_weights' in msa_module_kwargs['opm_kwargs'] else False
        self.return_attn_weights = msa_module_kwargs['pwa_kwargs']['return_attn_weights'] if 'return_attn_weights' in msa_module_kwargs['pwa_kwargs'] else False

    @property
    def device(self):
        """Device of the model."""
        return self.zero.device

    def turn_off_seq_attn(self):
        self.msa_module.turn_off_seq_attn()

    def turn_on_seq_attn(self):
        self.msa_module.turn_on_seq_attn()

    def forward(
        self,
        additional_molecule_feats,
        msa: Float['b s n d'], 
        prev_msa_repr: Float['b s n d'] | None = None,
        prev_pairwise_repr: Float['b n n d'] | None = None,
        recycled_input: bool = False,
        mask: Bool['b n'] | None = None,
        msa_mask: Bool['b s'] | None = None,
        full_mask: Bool['b s n'] | None = None,
        pairwise_mask: Bool['b n n'] | None = None,
        seq_weights: Float['b s'] | None = None
    ):      
        ########## Prepare pairwise representation ##########
        # Initialize pair representation (INIT WITH POSITIONAL EMBEDDING)
        pairwise_repr = self.relative_position_encoding(additional_molecule_feats = additional_molecule_feats)
        seq_len = msa.shape[2]
        seq_arange = torch.arange(seq_len, device = msa.device)
        token_bonds = einx.subtract('i, j -> i j', seq_arange, seq_arange).abs() == 1
        token_bonds_feats = self.token_bond_to_pairwise_feat(token_bonds.float())
        pairwise_repr = pairwise_repr + token_bonds_feats
        
        ########## Prepare MSA ##########
        # Initial MSA projection to representation dimensionality
        msa = self.msa_init_proj(msa)

        ########## Add previous iteration's MSA and pairwise representations ##########
        if recycled_input:
            assert prev_msa_repr is not None and prev_pairwise_repr is not None, "prev_msa_repr and prev_pairwise_repr must be provided if recycled_input is True"
            # Normalize the MSA and pairwise representation
            msa += self.prep_recycle_msa(prev_msa_repr)
            pairwise_repr += self.prep_recycle_pairwise(prev_pairwise_repr)
        
        ########## Pass through MSA module ##########
        # Updates the MSA and pairwise representations
        msa, pairwise_repr, seq_weights_list, attn_gates_list, attn_weights_list = self.msa_module(
            pairwise_repr = pairwise_repr,
            msa = msa,
            mask = mask,
            msa_mask = msa_mask,
            full_mask = full_mask,
            pairwise_mask = pairwise_mask,
            seq_weights = seq_weights
        )
        # Initialize output dictionary
        res = {}

        ########## Predict amino acid identities at each position in the MSA ##########
        # Get logits via language modeling head
        if self.query_only:
            logits = self.lm_head(msa[:, 0:1, :, :])
        else:
            logits = self.lm_head(msa)
        res['logits'] = logits

        ########## Return representations ##########
        if self.return_msa_repr:
            res["msa_repr"] = msa
        if self.return_pairwise_repr:
            res["pairwise_repr"] = pairwise_repr
        if self.return_seq_weights:
            res["seq_weights"] = seq_weights_list
            res["attn_gates"] = attn_gates_list
        if self.return_attn_weights:
            res["attn_weights"] = attn_weights_list
        return res

class MSAContactModel(Module):
    """
    Loads a pre-trained MSA language model and predicts contact maps from MSAs
    """
    def __init__(
        self,
        pretrained_weights_path: str,
        pretrained_is_final: bool,
        msa_model_kwargs: dict,
        dim_contact_hidden: int = None,
        dim_contact_input: int = None,
        compiled_checkpoint: bool = False,
        logits_grad: bool = False,
        logreg_contact_head: bool = False,
        multi_layer_logreg_contact_head: bool = False,
        multi_layer_logreg_contact_head_num_layers: int = None,
        apply_apc: bool = False
    ):
        super().__init__()
        self.msa_model_kwargs = msa_model_kwargs
        self.logits_grad = logits_grad
        if 'return_pairwise_repr' in msa_model_kwargs:
            assert msa_model_kwargs['return_pairwise_repr'] == True, "return_pairwise_repr must be True"
        self.pretrained_model = AF3MSAMLM(
            **msa_model_kwargs
        )
        checkpoint = torch.load(pretrained_weights_path, weights_only=True)
        if not pretrained_is_final:
            checkpoint = checkpoint['model_state_dict']
        if compiled_checkpoint:
            checkpoint = remove_orig_mod_prefix(checkpoint)

        missing_keys, unexpected_keys = self.pretrained_model.load_state_dict(checkpoint, strict=False)
        print("Missing keys: ", np.unique(['.'.join(k.split('.')[-2:]) for k in missing_keys]))
        print("Unexpected keys: ", unexpected_keys)
        self.pretrained_model.eval()
        # Initialize contact prediction head
        dim_input = dim_contact_input if dim_contact_input is not None else msa_model_kwargs['dim_pairwise']
        if logreg_contact_head:
            self.contact_head = LogisticRegressionContactHead(
                dim_pairwise = dim_contact_input,
            )
        elif multi_layer_logreg_contact_head:
            assert multi_layer_logreg_contact_head_num_layers is not None, "multi_layer_logreg_contact_head_num_layers must be provided if multi_layer_logreg_contact_head is True"
            self.contact_head = MultiLayerLogisticRegressionContactHead(
                dim_pairwise = dim_input,
                num_layers = multi_layer_logreg_contact_head_num_layers
            )
        else:
            assert dim_contact_hidden is not None, "dim_contact_hidden must be provided if logreg_contact_head is False"
            self.contact_head = ContactHead(
                dim_pairwise = dim_input,
                dim_dense_hidden = dim_contact_hidden
            )

    def turn_off_seq_attn(self):
        self.pretrained_model.turn_off_seq_attn()

    def turn_on_seq_attn(self):
        self.pretrained_model.turn_on_seq_attn()

    def forward(
        self,
        additional_molecule_feats,
        msa: Float['b s n d'], 
        mask: Bool['b n'] | None = None,
        msa_mask: Bool['b s'] | None = None,
        full_mask: Bool['b s n'] | None = None,
        pairwise_mask: Bool['b n n'] | None = None,
        seq_weights: Float['b s'] | None = None,
        use_layer_l: List[int] | None = None
    ):
        if not self.logits_grad:
            with torch.no_grad():
                res = self.pretrained_model(
                    additional_molecule_feats = additional_molecule_feats,
                    msa = msa,
                    mask = mask,
                    msa_mask = msa_mask,
                    full_mask = full_mask,
                    pairwise_mask = pairwise_mask,
                    seq_weights = seq_weights
                )
        else:
            res = self.pretrained_model(
                additional_molecule_feats = additional_molecule_feats,
                msa = msa,
                mask = mask,
                msa_mask = msa_mask,
                full_mask = full_mask,
                pairwise_mask = pairwise_mask,
                seq_weights = seq_weights
            )
        pairwise_repr = res['pairwise_repr']
        if use_layer_l is not None:
            assert pairwise_repr.dim() == 5, "pairwise_repr must be 5D if use_layer_l is provided (b, l, n, n, d)"
            b, l, n, _, d = pairwise_repr.shape
            use_nlayers = len(use_layer_l)
            # pairwise_repr = pairwise_repr[:, use_layer_l, :, :, :].view(b, n, n, d * use_nlayers)
            pairwise_repr = pairwise_repr[:, use_layer_l, :, :, :]
        predicted_contact_map = self.contact_head(pairwise_repr)
        res['predicted_contact_map'] = predicted_contact_map
        return res

    def compute_pairwise_repr(
        self,
        additional_molecule_feats,
        msa: Float['b s n d'], 
        mask: Bool['b n'] | None = None,
        msa_mask: Bool['b s'] | None = None,
        full_mask: Bool['b s n'] | None = None,
        pairwise_mask: Bool['b n n'] | None = None,
        seq_weights: Float['b s'] | None = None
    ):
        with torch.no_grad():
            res = self.pretrained_model(
                additional_molecule_feats = additional_molecule_feats,
                msa = msa,
                mask = mask,
                msa_mask = msa_mask,
                full_mask = full_mask,
                pairwise_mask = pairwise_mask,
                seq_weights = seq_weights
            )
        return res['pairwise_repr']

    def compute_contacts_from_pairwise_repr(
        self,
        pairwise_repr: Float['b n n d']
    ):
        return self.contact_head(pairwise_repr)

def remove_orig_mod_prefix(state_dict):
    return {k.replace('_orig_mod.', ''): v for k, v in state_dict.items()}


import torch
import torch.nn.functional as F
from torch.nn import Module
from torch import nn
from torch.nn import Linear, Sequential, LayerNorm
from torch import Tensor
from functools import partial
from typing import Literal
from einops import pack, unpack
import einx
from .custom_typing import (
    Float,
    Bool,
    typecheck
)

# Linear layer without bias term (simple linear transformation)
LinearNoBias = partial(Linear, bias = False)

class RMSNorm(Module):
    def __init__(self, dim: int, eps: float = 1e-6):
        super().__init__()
        self.eps = eps
        self.weight = nn.Parameter(torch.ones(dim))

    def _norm(self, x: Tensor) -> Tensor:
        return x * torch.rsqrt(x.pow(2).mean(-1, keepdim=True) + self.eps)

    def forward(self, x: Tensor) -> Tensor:
        output = self._norm(x.float()).type_as(x)
        return output * self.weight

class SwiGLU(Module):
    """
    Input is a tensor representing the projected output of previous layer and divides it 
    into two equal dimensionality tensors. One half is passed through Swish activation 
    and the output of this and the other half is multiplied element-wise
    """
    @typecheck
    def forward(
        self,
        x: Float['... d']
    ) -> Float['... d//2']:
        x, gates = x.chunk(2, dim=-1)
        return F.silu(gates) * x
        
class Transition(Module):
    """
    Algorithm 11
    x <- LayerNorm(x); x in R^{c}
    a = LinearNoBias(x); a in R^{n*c}
    b = LinearNoBias(x); b in R^{n*c}
    x <- LinearNoBias(swish(a) (*)  b); x in R^{c}
    """
    def __init__(
        self,
        dim,
        expansion_factor = 4
    ):
        super().__init__()
        
        dim_inner = int(dim * expansion_factor)
        # We apply PreLayerNorm to these transition layers so it's okay not to do it here
        self.ff = Sequential(
            LinearNoBias(dim, dim_inner * 2),
            SwiGLU(),
            LinearNoBias(dim_inner, dim)
        )

    @typecheck
    def forward(
        self,
        x: Float['... d'],
    ) -> Float['... d']:
        return self.ff(x)

class Dropout(Module):
    @typecheck
    def __init__(
        self,
        prob: float,
        dropout_type: Literal['row','col'] | None = None
    ):
        super().__init__()
        self.dropout = nn.Dropout(prob)
        self.dropout_type = dropout_type

    def forward(
        self,
        t: Tensor
    ) -> Tensor:
        # Check tensor dimensionality
        if self.dropout_type in {'row', 'col'}:
            assert t.ndim == 4, 'Tensor must be 4 dimensions for row / col structured dropout'
        # If dropout type is not specified, use standard dropout
        if not exists(self.dropout_type):
            return self.dropout(t)

        # Row (sequence) dropout
        if self.dropout_type == 'row':
            batch, row, _, _ = t.shape
            ones_shape = (batch, row, 1, 1)
        # Column (residue) dropout
        elif self.dropout_type == 'col':
            batch, _, col, _ = t.shape
            ones_shape = (batch, 1, col, 1)

        ones = t.new_ones(ones_shape)
        dropped = self.dropout(ones)
        return t * dropped

def exists(v):
    return v is not None

def default(v, d):
    return v if exists(v) else d

def pack_one(t, pattern):
    packed, ps = pack([t], pattern)

    def unpack_one(to_unpack, unpack_pattern = None):
        unpacked, = unpack(to_unpack, ps, default(unpack_pattern, pattern))
        return unpacked

    return packed, unpack_one

def to_pairwise_mask(
    mask_i: Bool['... n'],
    mask_j: Bool['... n'] | None = None
) -> Bool['... n n']:

    mask_j = default(mask_j, mask_i)
    assert mask_i.shape == mask_j.shape
    return einx.logical_and('... i, ... j -> ... i j', mask_i, mask_j)

def max_neg_value(t: Tensor):
    return -torch.finfo(t.dtype).max

class PreRMSNorm(Module):
    @typecheck
    def __init__(
        self,
        fn: Module,
        dim: int
    ):
        super().__init__()
        self.fn = fn
        self.norm = RMSNorm(dim)
    
    @typecheck
    def forward(
        self,
        x: Float['... n d'],
        **kwargs
    ) -> Float['... n d']:
        x = self.norm(x)
        return self.fn(x, **kwargs)

class PreLayerNorm(Module):
    @typecheck
    def __init__(
        self,
        fn: Module,
        dim: int
    ):
        super().__init__()
        self.fn = fn
        self.norm = LayerNorm(dim)

    @typecheck
    def forward(
        self,
        x: Float['... n d'],
        **kwargs
    ) -> Float['... n d']:
        # LayerNorm the input tensor
        x = self.norm(x)
        # Forward pass through specified function
        return self.fn(x, **kwargs)


import subprocess
import os
import torch
import multiprocessing as mp
from tqdm import tqdm
from subprocess import Popen, PIPE
from Bio.PDB import *
from typing import Optional, Dict
import matplotlib.pyplot as plt
from sklearn.mixture import GaussianMixture
import numpy as np
import torch.nn.functional as F

def _run(x):
    "Generic run."
    if isinstance(x, str):
        res = subprocess.run(x.split(' '), stdout=PIPE, stderr=PIPE, universal_newlines=True)
    elif isinstance(x, list):
        res = subprocess.run(x, stdout=PIPE, stderr=PIPE, universal_newlines=True)
    else:
        print("Must pass a string or a list of strings to _run()")
        return -1
    res.stdout = res.stdout.strip().split('\n')
    res.stderr = res.stderr.strip().split('\n')
    return res

def write_log(fout, res):
    """Write output logs."""
    with open(fout+'.out', "a") as f:
        for l in res.stdout:
            f.write(l+'\n')

    with open(fout+'.err', "a") as f:
        for l in res.stderr:
            f.write(l+'\n')

#############
# PDB utils #
#############
def write_chain_from_pdb(structure_file_path, chain_ids, out_file_path: None, ignore_hetatm = True):
    """Write chains from a PDB file."""
    atom_counter = 1
    atom_label_l = ['ATOM  ', 'HETATM'] if ignore_hetatm else ['ATOM  ']
    if out_file_path is None:
        chain_id_str = '_'.join(chain_ids)
        out_file_path = structure_file_path.replace('.pdb', f'_{chain_id_str}.pdb')
    with open(structure_file_path, "r") as inFile, open(out_file_path, "w") as outFile:
        for line in inFile.readlines():
            if line[:6] in ['HEADER', 'TITLE ']:
                outFile.write(line)
                continue
            line_chain = line[21]
            if line_chain not in chain_ids:
                continue
            if line[:6] in atom_label_l:
                new_line = (line[:6] + str(atom_counter).rjust(5) + line[11:])
                outFile.write(new_line)
                atom_counter += 1
            if line.startswith('TER'):
                new_ter = ("TER   " + str(atom_counter).rjust(5) + line[11:])
                outFile.write(new_ter)
            if line.startswith('END'):
                outFile.write(line)
    return out_file_path

def write_chain_from_cif(structure_file_path, chain_ids, out_file_path):
    # Create MMCIFParser
    parser = MMCIFParser(QUIET=True)
    # Parse structure from CIF file
    structure = parser.get_structure('protein_structure', structure_file_path)
    # Create subclass of Select to filter out chains
    class ChainSelect(Select):
        def __init__(self, chain_ids):
            self.chain_ids = chain_ids
        def accept_chain(self, chain):
            if chain.id in self.chain_ids:
                return True
            else:
                return False
    io = MMCIFIO()
    io.set_structure(structure)
    io.save(out_file_path, ChainSelect(chain_ids))

def convert_cif_to_pdb(input_cif_path, output_pdb_path):
    parser = MMCIFParser(QUIET=True)
    structure = parser.get_structure('protein_structure', input_cif_path)
    io = PDBIO()
    io.set_structure(structure)
    io.save(output_pdb_path)


def get_distance_matrix(structure_file_path, chain_id):
    # Load coordinates
    with open(structure_file_path, "r") as oFile:
        cif_lines_l = [l for l in oFile.readlines() if 
                       (l.startswith('ATOM') and (l.split()[6] == chain_id) and th
                        (((l.split()[3] == 'CA') and (l.split()[5]=='GLY')) or 
                         ((l.split()[3]=='CB') and (l.split()[5] != 'GLY'))))]
    coords_a = np.array([[float(xyz) for xyz in l.split()[10:13]] for l in cif_lines_l])
    # Compute distances
    dist_a = np.linalg.norm(coords_a[:, None] - coords_a, axis=-1)
    return dist_a


####################
# Contact analysis #
####################
def get_contacts(structure_file_path, chain_id, max_dist = 8):
    # Get distance matrix
    dist_a = get_distance_matrix(structure_file_path, chain_id)
    # Threshold distances to determine contacts
    contacts_a = dist_a.copy() <= max_dist
    return contacts_a

def compute_precision(predictions_a, targets_a, min_seq_sep = 6):
    assert predictions_a.shape == targets_a.shape, "Predictions and targets must have the same shape"
    # Get valid indices
    seqlen = predictions_a.shape[1]
    seqlen_range = torch.arange(seqlen)
    valid_mask = seqlen_range[None, :] - seqlen_range[:, None] >= min_seq_sep
    # Some contact maps have -1 for invalid pairs
    valid_mask = valid_mask & (targets_a >= 0)

def run_confind(
    structure_file_path, 
    output_contact_file_path, 
    bin_path = "/home/ubuntu/tools/confind/confind",
    rot_lib_path = "/home/ubuntu/tools/confind/rotlibs",
    log = None
):
    """Run confind."""
    cmd = f"{bin_path} --p {structure_file_path} --o {output_contact_file_path} --rLib {rot_lib_path}"
    res = _run(cmd)
    if log is not None:
        assert isinstance(log, str), "Log must be a string"
        write_log(log, res)
    return res

def run_confind_mp(param_d):
    """Run confind in parallel."""
    structure_file_path = param_d["structure_file_path"]
    output_contact_file_path = param_d["output_contact_file_path"]
    bin_path = param_d["bin_path"]
    rot_lib_path = param_d["rot_lib_path"]  
    run_confind(structure_file_path, output_contact_file_path, bin_path, rot_lib_path)

def run_batch_confind(
    structure_file_paths_l: list, 
    output_contact_file_paths_l: list, 
    bin_path = "/home/ubuntu/tools/confind/confind", 
    rot_lib_path = "/home/ubuntu/tools/confind/rotlibs",
    nproc = None,
    cpu_buffer: int = 4
):
    """Run confind in batch."""
    if nproc is None:
        nproc = mp.cpu_count() - cpu_buffer
    zipped_paths = zip(structure_file_paths_l, output_contact_file_paths_l)
    param_d_l = [{'structure_file_path': structure_file_path, 'output_contact_file_path': output_contact_file_path, 'bin_path': bin_path, 'rot_lib_path': rot_lib_path} for 
                 structure_file_path, output_contact_file_path in zipped_paths]
    with mp.Pool(processes=nproc) as pool:
        list(tqdm(pool.imap(run_confind_mp, param_d_l), total=len(param_d_l), leave=True))
        pool.close()


# ######################
# # Contact evaluation #
# ######################
# def calculate_precision_batch(pred_contacts, true_contacts, minsep=6, maxsep=None):
#     if pred_contacts.shape != true_contacts.shape:
#         raise ValueError("Predicted and true contact matrices must have the same shape")
#     if pred_contacts.dim() == 2:
#         pred_contacts = pred_contacts.unsqueeze(0)
#     if true_contacts.dim() == 2:
#         true_contacts = true_contacts.unsqueeze(0)
#     # Get batch size and longest sequence length
#     B, L, _ = pred_contacts.shape
#     # Get device
#     device = pred_contacts.device
#     if true_contacts.device != device:
#         true_contacts = true_contacts.to(device)
#     # Create valid mask (only consider upper triangular matrix)
#     # Padded regions have -1 in true_contacts
#     # Valid mask for both predicted and true contact matrices
#     seqlen_range = torch.arange(L, device=device)
#     sep = seqlen_range.unsqueeze(0) - seqlen_range.unsqueeze(1)
#     sep = sep.unsqueeze(0)
#     valid_mask = sep >= minsep
#     if maxsep is not None:
#         valid_mask &= sep < maxsep
#     valid_mask = valid_mask & (true_contacts >= 0)
    
#     # Fill prediction matrix with -inf if it's not valid
#     pred_contacts = pred_contacts.masked_fill(~valid_mask, float('-inf'))

#     # Get upper triangular predictions and true contacts
#     # Predictions have been masked with -inf if they are not valid, so when we sort and take the topk, we are only taking valid predictions
#     x_ind, y_ind = np.triu_indices(L, minsep)
#     predictions_upper = pred_contacts[:, x_ind, y_ind]
#     true_upper = true_contacts[:, x_ind, y_ind]
    
#     # Determine number of true contacts in valid region
#     K = true_contacts.masked_fill(~valid_mask, 0).sum(dim=-1).sum(dim=-1)
    
#     # Get Top-K predictions and pad if not enough predictions
#     max_k = int(K.max().item())
#     if max_k == 0:
#         return None
#     indices = predictions_upper.argsort(dim=-1, descending=True)[:, :max_k]
#     topk_targets = true_upper[torch.arange(B).unsqueeze(1), indices]
#     if topk_targets.size(1) < max_k:
#         topk_targets = torch.nn.functional.pad(topk_targets, [0, max_k - topk_targets.size(1)])
#     # Get cumulative sum of true positives
#     cumulative_dist = topk_targets.type_as(pred_contacts).cumsum(dim=-1)
#     # Compute precision based on true number of contacts
#     gather_lengths = K.unsqueeze(1)
#     gather_indices = (
#         torch.arange(0.1, 1.1, 0.1, device=device).unsqueeze(0) * gather_lengths
#     ).type(torch.long) - 1
#     gather_indices = gather_indices.clamp_min(0)
#     # Bin cumulative sum of true positives with intervals of sequence length
#     binned_cumulative_dist = cumulative_dist.gather(1, gather_indices)
#     binned_precisions = binned_cumulative_dist / (gather_indices + 1).type_as(binned_cumulative_dist)
#     # Get precisions 
#     pl5 = binned_precisions[:, 1]
#     pl2 = binned_precisions[:, 4]
#     pl = binned_precisions[:, 9]
#     auc = binned_precisions.mean(dim=-1)

#     return {"AUC": auc, "P@L5": pl5, "P@L2": pl2, "P@L": pl}

# def evaluate_contact_prediction(
#     predictions: torch.Tensor,
#     targets: torch.Tensor,
# ) -> Dict[str, float]:
#     if isinstance(targets, np.ndarray):
#         targets = torch.from_numpy(targets)
#     if isinstance(predictions, np.ndarray):
#         predictions = torch.from_numpy(predictions)
#     contact_ranges = [
#         ("local", 3, 6),
#         ("short", 6, 12),
#         ("medium", 12, 24),
#         ("long", 24, None),
#     ]
#     metrics = {}
#     targets = targets.to(predictions.device)
#     for name, minsep, maxsep in contact_ranges:
#         rangemetrics = calculate_precision_batch(
#             predictions,
#             targets,
#             minsep=minsep,
#             maxsep=maxsep,
#         )
#         if rangemetrics is not None:
#             for key, val in rangemetrics.items():
#                 metrics[f"{name}_{key}"] = list(val.float().cpu().numpy())
#     return metrics

def compute_precisions(
    predictions: torch.Tensor,
    targets: torch.Tensor,
    src_lengths: Optional[torch.Tensor] = None,
    minsep: int = 6,
    maxsep: Optional[int] = None,
    override_length: Optional[int] = None,  # for casp
):
    if isinstance(predictions, np.ndarray):
        predictions = torch.from_numpy(predictions)
    if isinstance(targets, np.ndarray):
        targets = torch.from_numpy(targets)
    if predictions.dim() == 2:
        predictions = predictions.unsqueeze(0)
    if targets.dim() == 2:
        targets = targets.unsqueeze(0)
    override_length = (targets[0, 0] >= 0).sum()

    # Check sizes
    if predictions.size() != targets.size():
        raise ValueError(
            f"Size mismatch. Received predictions of size {predictions.size()}, "
            f"targets of size {targets.size()}"
        )
    device = predictions.device

    batch_size, seqlen, _ = predictions.size()
    seqlen_range = torch.arange(seqlen, device=device)

    sep = seqlen_range.unsqueeze(0) - seqlen_range.unsqueeze(1)
    sep = sep.unsqueeze(0)
    valid_mask = sep >= minsep
    valid_mask = valid_mask & (targets >= 0)  # negative targets are invalid

    if maxsep is not None:
        valid_mask &= sep < maxsep

    if src_lengths is not None:
        valid = seqlen_range.unsqueeze(0) < src_lengths.unsqueeze(1)
        valid_mask &= valid.unsqueeze(1) & valid.unsqueeze(2)
    else:
        src_lengths = torch.full([batch_size], seqlen, device=device, dtype=torch.long)

    predictions = predictions.masked_fill(~valid_mask, float("-inf"))

    x_ind, y_ind = np.triu_indices(seqlen, minsep)
    predictions_upper = predictions[:, x_ind, y_ind]
    targets_upper = targets[:, x_ind, y_ind]

    topk = seqlen if override_length is None else max(seqlen, override_length)
    indices = predictions_upper.argsort(dim=-1, descending=True)[:, :topk]
    topk_targets = targets_upper[torch.arange(batch_size).unsqueeze(1), indices]
    if topk_targets.size(1) < topk:
        topk_targets = F.pad(topk_targets, [0, topk - topk_targets.size(1)])

    cumulative_dist = topk_targets.type_as(predictions).cumsum(-1)

    gather_lengths = src_lengths.unsqueeze(1)
    if override_length is not None:
        gather_lengths = override_length * torch.ones_like(
            gather_lengths, device=device
        )

    gather_indices = (
        torch.arange(0.1, 1.1, 0.1, device=device).unsqueeze(0) * gather_lengths
    ).type(torch.long) - 1

    binned_cumulative_dist = cumulative_dist.gather(1, gather_indices)
    binned_precisions = binned_cumulative_dist / (gather_indices + 1).type_as(
        binned_cumulative_dist
    )

    pl5 = binned_precisions[:, 1]
    pl2 = binned_precisions[:, 4]
    pl = binned_precisions[:, 9]
    auc = binned_precisions.mean(-1)

    return {"AUC": auc, "P@L": pl, "P@L2": pl2, "P@L5": pl5}


def evaluate_contact_prediction(
    predictions: torch.Tensor,
    targets: torch.Tensor,
) -> Dict[str, float]:
    if isinstance(targets, np.ndarray):
        targets = torch.from_numpy(targets)
    contact_ranges = [
        ("local", 3, 6),
        ("short", 6, 12),
        ("medium", 12, 24),
        ("long", 24, None),
        ("morethansix", 6, None)
    ]
    metrics = {}
    targets = targets.to(predictions.device)
    for name, minsep, maxsep in contact_ranges:
        rangemetrics = compute_precisions(
            predictions,
            targets,
            minsep=minsep,
            maxsep=maxsep,
        )
        for key, val in rangemetrics.items():
            metrics[f"{name}_{key}"] = val.item()
    return metrics

def extract_confind_contacts(confind_file_path):
    with open(confind_file_path, "r") as oFile:
        lines = oFile.readlines()
        # Get protein length
        max_res_idx = int(lines[-2].split()[1].split(',')[1])
        length = len(lines[-1].split()[1:])
        contact_a = np.zeros((max_res_idx, max_res_idx))
        for line in lines:
            if line.startswith('contact'):
                split_line = line.split()
                res_i = int(split_line[1].split(',')[1]) - 1
                res_j = int(split_line[2].split(',')[1]) - 1
                c = float(split_line[3])
                if c > contact_a[res_i, res_j]:
                    contact_a[res_i, res_j] = c
                    contact_a[res_j, res_i] = c
    return contact_a

def extract_homooligomeric_confind_contacts(confind_file_path):
    with open(confind_file_path, "r") as oFile:
        lines = oFile.readlines()
        # Get protein length
        observed_chains_d = {}
        for line in lines:
            if line.startswith("freedom"):
                chain_id = line.split()[1].split(',')[0]
                res_idx = int(line.split()[1].split(',')[1])
                if chain_id not in observed_chains_d:
                    observed_chains_d[chain_id] = res_idx
                else:
                    if res_idx > observed_chains_d[chain_id]:
                        observed_chains_d[chain_id] = res_idx
        max_res_idx = max(list(observed_chains_d.values()))
        contact_a = np.zeros((max_res_idx, max_res_idx))
        # Get contacts
        for line in lines:
            if line.startswith('contact'):
                split_line = line.split()
                res_i = int(split_line[1].split(',')[1]) - 1
                res_j = int(split_line[2].split(',')[1]) - 1
                c = float(split_line[3])
                if c > contact_a[res_i, res_j]:
                    contact_a[res_i, res_j] = c
                    contact_a[res_j, res_i] = c
    return contact_a

def fit_seq_weight_mixture_model(
    seq_weights_a: np.array,
    n_components: int = 2,
    random_state: int = 42,
    max_iter: int = 5000,
    n_init: int = 100,
    tol: float = 1e-6,
    scaling_factor = 1e5,
    plot = True,
    figsize = (6, 4),
    return_gmm = False
):
    # Fit a mixture model to the sequence weights
    # Prepare the data
    X = seq_weights_a.reshape(-1, 1)
    # Initialize and fit GMM to scaled data
    gmm = GaussianMixture(n_components=n_components, random_state=random_state, max_iter=max_iter, n_init=n_init, tol=tol)
    gmm.fit(X * scaling_factor)
    average_log_likelihood = gmm.score(X * scaling_factor)
    n_samples = X.shape[0]
    total_log_likelihood = average_log_likelihood * n_samples
    # Generate x-values in original scale
    x = np.linspace(seq_weights_a.min(), seq_weights_a.max(), 1000).reshape(-1, 1)
    # Compute PDFs in scaled space
    x_scaled = x * scaling_factor
    log_prob_scaled = gmm.score_samples(x_scaled)
    pdf_scaled = np.exp(log_prob_scaled)
    responsibilities = gmm.predict_proba(x_scaled)
    pdf_individual_scaled = responsibilities * pdf_scaled[:, np.newaxis]
    
    if plot:
        f, ax = plt.subplots(1, 1, figsize=figsize)
        
        # Plot histogram of original data with density=False to show counts
        hist_counts, hist_bins, _ = ax.hist(seq_weights_a, bins=100, density=False, 
                                          alpha=0.5, color="grey", 
                                          label="Avg. sequence attention weights")
        
        # Calculate bin width for scaling PDFs to match count data
        bin_width = hist_bins[1] - hist_bins[0]
        
        # Scale factor to convert density to counts
        scale_to_counts = bin_width * len(seq_weights_a)
        
        # Plot the PDFs scaled to match counts
        for component_idx in range(n_components):
            # Transform PDFs to count scale
            pdf_component = pdf_individual_scaled[:, component_idx] * scaling_factor * scale_to_counts
            ax.plot(x, pdf_component, '--', color=f'C{component_idx}', label=f"Component {component_idx}")
        
        # Set labels and title
        ax.set_xlabel("Sequence weight", size=12)
        ax.set_ylabel("Count", size=12)  # Changed from "Density" to "Count"
        ax.set_title("OmpR sequence weight distribution", size=12)  # Updated title to match your image
        
        # Add vertical line at x = 1/num_seqs
        ax.axvline(x=1 / seq_weights_a.shape[0], color='grey', linestyle='--', label="Uniform weighting")
        ax.legend(fontsize=10)
        
        return gmm.means_ / scaling_factor, gmm.covariances_ / scaling_factor**2, total_log_likelihood, f, ax
    
    return gmm.means_ / scaling_factor, gmm.covariances_ / scaling_factor**2, total_log_likelihood


# Get top L pairs
def get_top_L_pairs(predicted_contacts_a, minsep=24):
    L = predicted_contacts_a.shape[0]
    # Only consider upper triangular part
    triu_idx = np.triu_indices_from(predicted_contacts_a, 1)
    # Mask out pairs < 24 residues apart
    mask = np.abs(triu_idx[0] - triu_idx[1]) >= minsep
    filtered_i = triu_idx[0][mask]
    filtered_j = triu_idx[1][mask]
    # Get values for these filtered indices
    vals = predicted_contacts_a[filtered_i, filtered_j]
    # Get top L
    L = min(L, len(vals))
    cutoff = np.sort(vals)[::-1][L-1]
    vals_sort_idx = vals >= cutoff
    return filtered_i[vals_sort_idx], filtered_j[vals_sort_idx]


# Adapted from https://github.com/yoakiyama/openfold/blob/main/openfold/utils/chunk_utils.py

import torch
from typing import Callable, Dict, Optional, Tuple, Any
from functools import partial

def dict_map(fn, dic, leaf_type):
    new_dict = {}
    for k, v in dic.items():
        if type(v) is dict:
            new_dict[k] = dict_map(fn, v, leaf_type)
        else:
            new_dict[k] = tree_map(fn, v, leaf_type)

    return new_dict

def spec_dict_map(fn_d, dic, leaf_type):
    new_dict = {}
    for k, v in dic.items():
        if type(v) is dict:
            new_dict[k] = dict_map(fn_d[k], v, leaf_type)
        else:
            new_dict[k] = tree_map(fn_d[k], v, leaf_type)

    return new_dict

def tree_map(fn, tree, leaf_type):
    if isinstance(tree, dict) and isinstance(fn, dict):
        return spec_dict_map(fn, tree, leaf_type)
    elif isinstance(tree, dict):
        return dict_map(fn, tree, leaf_type)
    elif isinstance(tree, list):
        return [tree_map(fn, x, leaf_type) for x in tree]
    elif isinstance(tree, tuple):
        return tuple([tree_map(fn, x, leaf_type) for x in tree])
    elif isinstance(tree, leaf_type):
        return fn(tree)
    else:
        raise ValueError(f"Tree of type {type(tree)} not supported")
tensor_tree_map = partial(tree_map, leaf_type=torch.Tensor)

def _fetch_dims(tree):
    shapes = []
    tree_type = type(tree)
    if tree_type is dict:
        for v in tree.values():
            shapes.extend(_fetch_dims(v))
    elif tree_type is list or tree_type is tuple:
        for t in tree:
            shapes.extend(_fetch_dims(t))
    elif tree_type is torch.Tensor:
        shapes.append(tree.shape)
    else:
        raise ValueError("Not supported")

    return shapes

def chunk_layer(
    layer: Callable,
    inputs: Dict[str, Any],
    chunk_size: int,
    no_batch_dims: int,
    low_mem: bool = False,
    _out: Any = None,
    _add_into_out: bool = False,
    select_chunk_fn_d: Optional[Dict[str, Callable]] = None,
    orig_batch_dims: Optional[Dict[str, Tuple]] = None,
    flat_batch_dim: Optional[int] = None,
    og_batch_dim: Optional[Tuple] = None,
) -> Any:
    if not (len(inputs) > 0):
        raise ValueError("Must provide at least one input")

    if orig_batch_dims is None:
        orig_batch_dims = {}
        initial_dims = [shape[:no_batch_dims] for shape in _fetch_dims(inputs)]
        for k in inputs.keys():
            orig_batch_dims[k] = tuple([max(s) for s in zip(*initial_dims)])
        og_batch_dim = tuple([max(s) for s in zip(*initial_dims)])

    def _prep_inputs(t, batch_dim):
            if(not low_mem):
                if not sum(t.shape[:no_batch_dims]) == no_batch_dims:
                    t = t.expand(batch_dim + t.shape[no_batch_dims:])
                t = t.reshape(-1, *t.shape[no_batch_dims:])
            else:
                t = t.expand(batch_dim + t.shape[no_batch_dims:])
            return t
    prep_inputs_fn_d = {}
    for k in inputs.keys():
        prep_inputs_fn_d[k] = partial(_prep_inputs, batch_dim=orig_batch_dims[k])

    prepped_inputs = tensor_tree_map(prep_inputs_fn_d, inputs)
    prepped_outputs = None
    if(_out is not None):
        reshape_fn = lambda t: t.view([-1] + list(t.shape[no_batch_dims:]))
        prepped_outputs = tensor_tree_map(reshape_fn, _out)

    if flat_batch_dim is None:
        flat_batch_dim = 1
        for d in orig_batch_dims:
            flat_batch_dim *= orig_batch_dims[d][0]

    no_chunks = flat_batch_dim // chunk_size + (
        flat_batch_dim % chunk_size != 0
    )

    i = 0
    out = prepped_outputs
    for _ in range(no_chunks):
        # Chunk the input
        if(not low_mem):
            if select_chunk_fn_d is None:
                select_chunk = (
                    lambda t: t[i : i + chunk_size] if t.shape[0] != 1 else t
                )
            else:
                select_chunk = {k: partial(select_chunk_fn_d[k], i=i, chunk_size=chunk_size) for k in select_chunk_fn_d.keys()}
        else:
            select_chunk = (
                partial(
                    _chunk_slice, 
                    flat_start=i, 
                    flat_end=min(flat_batch_dim, i + chunk_size), 
                    no_batch_dims=len(orig_batch_dims)
                )
            )

        chunks = tensor_tree_map(select_chunk, prepped_inputs)

        # Run the layer on the chunk
        output_chunk = layer(**chunks)

        # Allocate space for the output
        if out is None:
            allocate = lambda t: t.new_zeros((flat_batch_dim,) + t.shape[1:])
            out = tensor_tree_map(allocate, output_chunk)

        # Put the chunk in its pre-allocated space
        out_type = type(output_chunk)
        if out_type is dict:
            def assign(d1, d2):
                for k, v in d1.items():
                    if type(v) is dict:
                        assign(v, d2[k])
                    else:
                        if(_add_into_out):
                            v[i: i + chunk_size] += d2[k]
                        else:
                            v[i: i + chunk_size] = d2[k]

            assign(out, output_chunk)
        elif out_type is tuple:
            for x1, x2 in zip(out, output_chunk):
                if(_add_into_out):
                    x1[i: i + chunk_size] += x2
                else:
                    x1[i : i + chunk_size] = x2
        elif out_type is torch.Tensor:
            if(_add_into_out):
                out[i: i + chunk_size] += output_chunk
            else:
                out[i: i + chunk_size] = output_chunk
        else:
            raise ValueError("Not supported")

        i += chunk_size

    reshape = lambda t: t.view(og_batch_dim + t.shape[1:])
    out = tensor_tree_map(reshape, out)

    return out


import torch
from torch.nn import Module, ModuleList, Sequential
import numpy as np
import einx
from math import exp
from functools import partial
from einops.layers.torch import Rearrange
from huggingface_hub import snapshot_download
from pathlib import Path
from typing import List

from MSA_Pairformer.core import LinearNoBias, PreLayerNorm, Transition, exists
from MSA_Pairformer.outer_product import OuterProduct
from MSA_Pairformer.regression import LMHead, LogisticRegressionContactHead
from MSA_Pairformer.pairwise_operations import MSAPairWeightedAveraging, PairwiseBlock
from MSA_Pairformer.positional_encoding import RelativePositionEncoding
from MSA_Pairformer.custom_typing import Float, Bool

class CoreModule(Module):
    """
    Core module for MSA Pairformer which includes stacked layers of:
    1) MSA pair weighted averaging (updates MSA representation using pairwise relationships from the pair representation)
    2) Query-biased outer product (updates pair representation using MSA representation)
    3) Triangle updates (updates pair representation using triplet information)
    """
    def __init__(
        self,
        *,
        depth: int = 22,
        dim_pairwise: int = 256,
        dim_msa: int = 464,
        opm_kwargs: dict = dict(
            dim_opm_hidden = 16,
            outer_product_flavor = "presoftmax_differential_attention",
            seq_attn = True,
            dim_qk = 128,
            chunk_size = None,
            return_seq_weights = True,
            lambda_init = None,
            eps = 1e-32,
        ),
        pwa_kwargs: dict = dict(
            heads = 8,
            dim_head = 32,
            dropout = 0.0,
            dropout_type = "row",
        ),
        pairwise_block_kwargs: dict = dict(
            dropout_row_prob = 0,
            dropout_col_prob = 0,
            tri_mult_dim_hidden = None, # Defaults to pair representation dimension
            use_triangle_updates = True,
            use_pair_updates = False # For ablation study
        )
    ):
        super().__init__()

        # Store parameters
        self.dim_pairwise = dim_pairwise
        self.depth = depth

        # Automatically assign lambda init if not provided (for presoftmax differential attention)
        if ('lambda_init' not in opm_kwargs) or (opm_kwargs['lambda_init'] is None):
            auto_lambda_init = True
        else:
            auto_lambda_init = False
            lambda_init = opm_kwargs['lambda_init']
        if 'lambda_init' in opm_kwargs:
            opm_kwargs.pop("lambda_init")

        # Initialize module stack
        layers = ModuleList([])
        for layer_idx in range(depth):
            curr_module_list = ModuleList()

            # MSA pair weighted averaging with gating -> transition
            msa_pair_weighted_avg = MSAPairWeightedAveraging(
                dim_msa = dim_msa, 
                dim_pairwise = dim_pairwise,
                **pwa_kwargs
            )
            msa_pre_ln = partial(PreLayerNorm, dim = dim_msa)
            curr_module_list.append(msa_pair_weighted_avg)
            curr_module_list.append(msa_pre_ln(Transition(dim = dim_msa)))

            # Outer product
            if auto_lambda_init and ('differential' in opm_kwargs['outer_product_flavor']):
                lambda_init = 0.8 - 0.6 * exp(-0.3 * layer_idx)
                lambda_init = torch.tensor(lambda_init, dtype=torch.bfloat16)
            else:
                lambda_init = None
            opm = OuterProduct(
                dim_msa = dim_msa,
                dim_pairwise = dim_pairwise,
                lambda_init = lambda_init,
                **opm_kwargs
            )
            curr_module_list.append(opm)

            # Pairwise representation update block
            pairwise_block = PairwiseBlock(
                dim_pairwise = dim_pairwise,
                **pairwise_block_kwargs
            )
            curr_module_list.append(pairwise_block)

            # Append all blocks of current layer to module list
            layers.append(curr_module_list)

        # Store layers in class object
        self.layers = layers

        # Final MSA update using MSA pair weighted averaging with gating
        self.final_msa_pwa = MSAPairWeightedAveraging(
            dim_msa = dim_msa, 
            dim_pairwise = dim_pairwise,
            **pwa_kwargs
        )
        # Transition module
        msa_pre_ln = partial(PreLayerNorm, dim = dim_msa)
        self.final_msa_transition = msa_pre_ln(Transition(dim = dim_msa))

        # Other parameters
        self.register_buffer('zero', torch.tensor(0.), persistent = False)

    def turn_off_query_biasing(self):
        for layer in self.layers:
            for module in layer:
                if isinstance(module, OuterProduct):
                    module.opm.seq_attn = False

    def turn_on_query_biasing(self):
        for layer in self.layers:
            for module in layer:
                if isinstance(module, OuterProduct):
                    module.opm.seq_attn = True

    def forward(
        self,
        msa: Float['b s n dm'],
        pairwise_repr: Float['b n n dp'],
        mask: Bool['b n'] | None = None, # Column mask
        msa_mask: Bool['b s'] | None = None, # Row mask,
        full_mask: Bool['b s n'] | None = None,
        pairwise_mask: Bool['b n n'] | None = None,
        seq_weights: Float['b s'] | None = None,
        seq_weights_dict: dict = None,
        query_only: bool = True,
        return_msa_repr_layer_idx: List[int] | int | None = None,
        return_pairwise_repr_layer_idx: List[int] | int | None = None,
        return_repr_after_layer_idx: int | None = None,
        return_seq_weights: bool = False,
    ) -> Float['b s n dm']:
        # Track seq weights, pairwise representations, and MSA representations for specified layers
        # seq weights are tracked throughout all layers
        seq_weights_list_d = {}
        pairwise_repr_d = {}
        msa_repr_d = {}
        if isinstance(return_msa_repr_layer_idx, int):
            return_msa_repr_layer_idx = [return_msa_repr_layer_idx]
        elif return_msa_repr_layer_idx is None:
            return_msa_repr_layer_idx = []
        if isinstance(return_pairwise_repr_layer_idx, int):
            return_pairwise_repr_layer_idx = [return_pairwise_repr_layer_idx]
        elif return_pairwise_repr_layer_idx is None:
            return_pairwise_repr_layer_idx = []
            
        # Pass MSA through each layer of the core module stack
        for layer_idx, (
            msa_pair_weighted_avg,
            msa_transition,
            outer_product,
            pairwise_block
        ) in enumerate(self.layers):
            # Pair weighted averaging (with residual connection)
            msa_residual = msa_pair_weighted_avg(msa = msa, pairwise_repr = pairwise_repr, mask = mask)
            msa = msa + msa_residual
            msa = msa + msa_transition(msa)
            if layer_idx in return_msa_repr_layer_idx:
                msa_repr_d[f"layer_{layer_idx}"] = msa[:, :1, :, :].cpu() if query_only else msa
            del msa_residual

            # Compute outer product mean (with residual connection)
            curr_seq_weights = None
            if seq_weights_dict is not None:
                if f"layer_{layer_idx}" in seq_weights_dict:
                    curr_seq_weights = seq_weights_dict[f"layer_{layer_idx}"].to(msa.device)
            elif seq_weights is not None:
                curr_seq_weights = seq_weights

            update_pairwise_repr, norm_weights = outer_product(
                msa = msa,
                mask = mask,
                msa_mask = msa_mask,
                full_mask = full_mask,
                pairwise_mask = pairwise_mask,
                seq_weights = curr_seq_weights
            )
            pairwise_repr = pairwise_repr + update_pairwise_repr
            del update_pairwise_repr
            if return_seq_weights:
                seq_weights_list_d[f"layer_{layer_idx}"] = norm_weights.cpu()

            # Pairwise representation block
            pairwise_repr = pairwise_block(pairwise_repr = pairwise_repr, mask = mask)
            if layer_idx in return_pairwise_repr_layer_idx:
                pairwise_repr_d[f"layer_{layer_idx}"] = pairwise_repr.cpu()

            # Break out of loop early if we've reached the layer from which we want to compute the representations
            if (return_repr_after_layer_idx is not None) and (layer_idx == return_repr_after_layer_idx):
                break
        layer_idx += 1

        # Final MSA update
        if (return_repr_after_layer_idx is None) or (layer_idx == return_repr_after_layer_idx):
            msa_residual = self.final_msa_pwa(
                msa = msa,
                pairwise_repr = pairwise_repr,
                mask = mask
            )
            msa = msa + msa_residual
            del msa_residual
            msa = msa + self.final_msa_transition(msa)
            if layer_idx in return_msa_repr_layer_idx:
                msa_repr_d[f"layer_{layer_idx+1}"] = msa[:, :1, :, :].cpu() if query_only else msa
        
        # Organize results
        res = {}
        res['final_msa_repr'] = msa[:, :1, :, :].cpu() if query_only else msa
        res['final_pairwise_repr'] = pairwise_repr.cpu()
        res['msa_repr_d'] = msa_repr_d
        res['pairwise_repr_d'] = pairwise_repr_d
        res['seq_weights_list_d'] = seq_weights_list_d
        return res

class MSAPairformer(Module):
    """
    3 main components:
        1) Core module
            a) Takes as input i) MSA representation; ii) Pair representation
            b) Iteratively updates MSA representation and pair representation bidirectionally
            c) Outputs refined MSA and pair representations
        2) MSA language model head
            b) Takes as input the final MSA representation
            c) Outputs the logits for the MSA language model
        3) Contact head
            a) takes as input a pairwise representation (in the final released model, this is the representation at layer 15)
            b) outputs a contact prediction between 0 and 1
    """
    def __init__(
        self,
        *,
        dim_msa_input: int = 28,
        dim_pairwise = 256,
        dim_msa = 464,
        dim_logits = 26,
        core_module_kwargs: dict = dict(
            depth = 22,
            opm_kwargs = dict(
                dim_opm_hidden = 16,
                outer_product_flavor = "presoftmax_differential_attention",
                seq_attn = True,
                dim_qk = 128,
                chunk_size = None,
                return_seq_weights = True,
                lambda_init = None,
                eps = 1e-32
            ),
            pwa_kwargs = dict(
                heads = 8,
                dim_head = 32,
                dropout = 0.0,
                dropout_type = "row",
            ),
            pairwise_block_kwargs = dict(
                dropout_row_prob = 0,
                dropout_col_prob = 0,
                tri_mult_dim_hidden = None,
                use_triangle_updates = True,
                use_pair_updates = False
            ),
        ),
        relative_position_encoding_kwargs: dict = dict(
            r_max = 32, 
            s_max = 2,
        ),
        contact_layer: int = 15,
        confind_contact_layer: int = 18
    ):
        super().__init__()
        self.dim_pairwise = dim_pairwise
        self.dim_msa = dim_msa
        self.contact_layer = contact_layer
        self.confind_contact_layer = confind_contact_layer

        # Relative position encoding
        self.relative_position_encoding = RelativePositionEncoding(
            dim_out = dim_pairwise,
            **relative_position_encoding_kwargs,
        )
        
        self.token_bond_to_pairwise_feat = Sequential(
            Rearrange('... -> ... 1'),
            LinearNoBias(1, dim_pairwise)
        )

        # Initial MSA projection
        self.msa_init_proj = LinearNoBias(dim_msa_input, dim_msa) if exists(dim_msa_input) else torch.nn.Identity()
        
        # Core module
        self.core_stack = CoreModule(
            dim_msa = dim_msa,
            dim_pairwise = dim_pairwise,
            **core_module_kwargs
        )

        # Projection layer for logits
        self.lm_head = LMHead(
            dim_msa,
            dim_logits
        )

        self.contact_head = LogisticRegressionContactHead(
            dim_pairwise = dim_pairwise,
        )
        self.confind_contact_head = LogisticRegressionContactHead(
            dim_pairwise = dim_pairwise
        )

    @property
    def device(self):
        """Device of the model."""
        return self.core_stack.zero.device

    ###### Load model ######
    @classmethod
    def from_pretrained(
        cls,
        device: torch.device | None = None,
        weights_dir: str | None = None,
    ):
        if device is None:
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        path = Path(snapshot_download(repo_id="yakiyama/MSA-Pairformer", cache_dir=weights_dir))
        checkpoint = torch.load(path / "model.bin", weights_only=True, map_location=device)
        confind_contact_checkpoint = torch.load(path / "confind_contact.bin", weights_only=True, map_location=device)
        cb_contact_checkpoint = torch.load(path / "contact.bin", weights_only=True, map_location=device)
        checkpoint.update(confind_contact_checkpoint)
        checkpoint.update(cb_contact_checkpoint)
        model = cls()
        model.load_state_dict(checkpoint)
        model.to(device)
        return model

    ###### Embed sequences ######
    def init_representations(
        self,
        msa: Float['b s n d'],
        complex_chain_break_indices: List[int] | None = None,
    ):
        # Initialize pair representation
        batch_size, _, seq_len, _ = msa.shape
        pairwise_repr = self.relative_position_encoding(
            batch_size = batch_size,
            seq_len = seq_len,
            device = msa.device,
            complex_chain_break_indices = complex_chain_break_indices
        )
        seq_arange = torch.arange(seq_len, device = msa.device)
        token_bonds = einx.subtract('i, j -> i j', seq_arange, seq_arange).abs() == 1
        token_bonds_feats = self.token_bond_to_pairwise_feat(token_bonds.float())
        pairwise_repr = pairwise_repr + token_bonds_feats

        # Initialize MSA representation
        msa = self.msa_init_proj(msa)

        return msa, pairwise_repr

    # Toggle query-biased attention
    def turn_off_query_biasing(self):
        self.core_stack.turn_off_query_biasing()

    def turn_on_query_biasing(self):
        self.core_stack.turn_on_query_biasing()

    ###### Make predictions / embeddings ######
    def forward(
        self,
        msa: Float['b s n d'],
        mask: Bool['b n'] | None = None,
        msa_mask: Bool['b s'] | None = None,
        full_mask: Bool['b s n'] | None = None,
        pairwise_mask: Bool['b n n'] | None = None,
        seq_weights: Float['b s'] | None = None,
        seq_weights_dict: dict = {},
        return_contacts: bool = True,
        return_seq_weights: bool = False,
        query_only: bool = True,
        return_pairwise_repr_layer_idx: List[int] | int | None = None,
        return_msa_repr_layer_idx: List[int] | int | None = None,
        complex_chain_break_indices: List[int] | None = None,
        return_repr_after_layer_idx: int | None = None,
    ):
        # Ensure that contact layer is in return_pairwise_repr_layer_idx if returning contacts
        if return_contacts:
            if return_pairwise_repr_layer_idx is None:
                return_pairwise_repr_layer_idx = [self.contact_layer, self.confind_contact_layer]
            elif isinstance(return_pairwise_repr_layer_idx, int):
                return_pairwise_repr_layer_idx = [return_pairwise_repr_layer_idx, self.contact_layer, self.confind_contact_layer]
            elif self.contact_layer not in return_pairwise_repr_layer_idx:
                return_pairwise_repr_layer_idx.append(self.contact_layer)
            if self.confind_contact_layer not in return_pairwise_repr_layer_idx:
                return_pairwise_repr_layer_idx.append(self.confind_contact_layer)

        # Initialize representations
        msa, pairwise_repr = self.init_representations(msa, complex_chain_break_indices)
        # Pass through layers
        results = self.core_stack(
            pairwise_repr = pairwise_repr,
            msa = msa,
            mask = mask,
            msa_mask = msa_mask,
            full_mask = full_mask,
            pairwise_mask = pairwise_mask,
            seq_weights = seq_weights,
            seq_weights_dict = seq_weights_dict,
            query_only = query_only,
            return_seq_weights = return_seq_weights,
            return_msa_repr_layer_idx = return_msa_repr_layer_idx,
            return_pairwise_repr_layer_idx = return_pairwise_repr_layer_idx,
            return_repr_after_layer_idx = return_repr_after_layer_idx
        )

        # Get logits via language modeling head
        if query_only:
            logits = self.lm_head(results['final_msa_repr'].to(self.device))
        else:
            logits = self.lm_head(results['final_msa_repr'].to(self.device))
        results['logits'] = logits

        # Predict contacts
        if return_contacts:
            contacts = self.contact_head(results['pairwise_repr_d'][f'layer_{self.contact_layer}'].to(self.device))
            results['predicted_cb_contacts'] = contacts
            confind_contacts = self.confind_contact_head(results['pairwise_repr_d'][f'layer_{self.confind_contact_layer}'].to(self.device))
            results['predicted_confind_contacts'] = confind_contacts
        return results

    ###### Contact prediction ######
    def predict_contacts(
        self,
        msa: Float['b s n d'],
        mask: Bool['b n'] | None = None,
        msa_mask: Bool['b s'] | None = None,
        full_mask: Bool['b s n'] | None = None,
        pairwise_mask: Bool['b n n'] | None = None,
        seq_weights: Float['b s'] | None = None,
        seq_weights_dict: dict = {},
        complex_chain_break_indices: List[int] | None = None,
        return_seq_weights: bool = False,
    ):
        # Initialize representations
        msa, pairwise_repr = self.init_representations(msa, complex_chain_break_indices)
        # Pass through layers
        results = self.core_stack(
            pairwise_repr = pairwise_repr,
            msa = msa,
            mask = mask,
            msa_mask = msa_mask,
            full_mask = full_mask,
            pairwise_mask = pairwise_mask,
            seq_weights = seq_weights,
            seq_weights_dict = seq_weights_dict,
            query_only = True,
            return_repr_after_layer_idx = None,
            return_pairwise_repr_layer_idx = [self.contact_layer, self.confind_contact_layer],
            return_seq_weights = return_seq_weights,
        )

        # Predict contacts
        cb_contacts = self.contact_head(results['pairwise_repr_d'][f'layer_{self.contact_layer}'].to(self.device))
        confind_contacts = self.confind_contact_head(results['pairwise_repr_d'][f'layer_{self.confind_contact_layer}'].to(self.device))
        res = {}
        res['predicted_cb_contacts'] = cb_contacts
        res['predicted_confind_contacts'] = confind_contacts
        if return_seq_weights:
            res['seq_weights_list_d'] = results['seq_weights_list_d']
        return res

    def predict_cb_contacts(
        self,
        msa: Float['b s n d'],
        mask: Bool['b n'] | None = None,
        msa_mask: Bool['b s'] | None = None,
        full_mask: Bool['b s n'] | None = None,
        pairwise_mask: Bool['b n n'] | None = None,
        seq_weights: Float['b s'] | None = None,
        seq_weights_dict: dict = {},
        complex_chain_break_indices: List[int] | None = None,
        return_seq_weights: bool = False,
    ):
        # Initialize representations
        msa, pairwise_repr = self.init_representations(msa, complex_chain_break_indices)
        # Pass through layers
        results = self.core_stack(
            pairwise_repr = pairwise_repr,
            msa = msa,
            mask = mask,
            msa_mask = msa_mask,
            full_mask = full_mask,
            pairwise_mask = pairwise_mask,
            seq_weights = seq_weights,
            seq_weights_dict = seq_weights_dict,
            query_only = True,
            return_repr_after_layer_idx = self.contact_layer,
            return_pairwise_repr_layer_idx = [self.contact_layer],
            return_seq_weights = return_seq_weights,
        )

        # Predict contacts
        contacts = self.contact_head(results['pairwise_repr_d'][f'layer_{self.contact_layer}'].to(self.device))
        res = {}
        res['predicted_cb_contacts'] = contacts
        if return_seq_weights:
            res['seq_weights_list_d'] = results['seq_weights_list_d']

        return res

    def predict_confind_contacts(
        self,
        msa: Float['b s n d'],
        mask: Bool['b n'] | None = None,
        msa_mask: Bool['b s'] | None = None,
        full_mask: Bool['b s n'] | None = None,
        pairwise_mask: Bool['b n n'] | None = None,
        seq_weights: Float['b s'] | None = None,
        seq_weights_dict: dict = {},
        complex_chain_break_indices: List[int] | None = None,
        return_seq_weights: bool = False,
    ):
        # Initialize representations
        msa, pairwise_repr = self.init_representations(msa, complex_chain_break_indices)
        # Pass through layers
        results = self.core_stack(
            pairwise_repr = pairwise_repr,
            msa = msa,
            mask = mask,
            msa_mask = msa_mask,
            full_mask = full_mask,
            pairwise_mask = pairwise_mask,
            seq_weights = seq_weights,
            seq_weights_dict = seq_weights_dict,
            query_only = True,
            return_repr_after_layer_idx = self.confind_contact_layer,
            return_pairwise_repr_layer_idx = [self.confind_contact_layer],
            return_seq_weights = return_seq_weights,
        )

        # Predict contacts
        contacts = self.confind_contact_head(results['pairwise_repr_d'][f'layer_{self.confind_contact_layer}'].to(self.device))
        res = {}
        res['predicted_confind_contacts'] = contacts
        if return_seq_weights:
            res['seq_weights_list_d'] = results['seq_weights_list_d']

        return res


import matplotlib.pyplot as plt
import numpy as np
from copy import deepcopy

def plot_contact_map(
    cons,
    L=1,
    minsep=24,
    cutoffs=[None],
    ss=[5],
    cc=["gray"],
    f=None,
    ax=None, 
    vmin=None, 
    vmax=None
):
    max_L = 0
    if ax is None:
        f, ax = plt.subplots(1, 1, figsize=(10, 10))

    #### Get upper triangle matrix for heatmap ####
    prot_len = cons[0].shape[0]
    upper = np.full((prot_len, prot_len), np.nan)
    for i in range(prot_len):
        for j in range(i, prot_len):  # i <= j
            if j - i >= minsep:
                upper[i, j] = cons[1][i, j]
            else:
                upper[i, j] = np.nan
    im = ax.imshow(upper, origin="upper", cmap="Greys", rasterized=True, vmin=vmin, vmax=vmax)

    #### Plot lower triangle matrix using top-L or ground truth contacts ####

    for n, (con, cutoff, s, c) in enumerate(zip(cons,cutoffs,ss,cc)):
        if con.shape[0] > max_L:
            max_L = con.shape[0]
        
        # Get upper triangular indices
        triu_idx = np.triu_indices_from(con, 1)
        
        # Create a mask for contacts that are at least minsep apart
        mask = np.abs(triu_idx[0] - triu_idx[1]) >= minsep
        
        # Apply the mask to the indices
        filtered_i = triu_idx[0][mask]
        filtered_j = triu_idx[1][mask]
        
        # Get values for these filtered indices
        vals = con[filtered_i, filtered_j]
        
        if cutoff is None:
            top = int(con.shape[0] * L)
            # Use the filtered indices directly
            if len(vals) > top:  # Make sure we have enough values
                cutoff = np.sort(vals)[::-1][top]
            else:
                cutoff = np.min(vals) if len(vals) > 0 else 0
        
        if np.all((con == 0) | (con == 1)):
            vals_sort_idx = vals >= cutoff
            label = "Ground truth"
        else:
            vals_sort_idx = vals > cutoff
            label = "True positives"
        
        # Use the filtered indices with the sorting
        if label == "Ground truth":
            alpha = 1
        else:
            alpha = 0.8
        if label == "Ground truth":
            # ax.scatter(filtered_i[vals_sort_idx], filtered_j[vals_sort_idx], c=c, s=s, rasterized=True, label=label)
            ax.scatter(filtered_i[vals_sort_idx], filtered_j[vals_sort_idx], c=c, s=s, rasterized=True, edgecolors='none')#, label=label)
            ax.scatter([-1], [-1], c=c, s=50, label=label, rasterized=True, edgecolors='none')
        
        if n == 1 and len(cons) > 1:
            # For the second contact map, highlight false positives (where ground truth is 0)
            # We need to match indices between the two contact maps
            if filtered_i.size > 0:  # Make sure we have pairs to check
                # Get values from first contact map at same positions
                first_map_vals = cons[0][filtered_i[vals_sort_idx], filtered_j[vals_sort_idx]]
                bad = first_map_vals == 0
                
                # Plot false positives
                # ax.scatter(filtered_i[vals_sort_idx][bad], filtered_j[vals_sort_idx][bad], c="red", s=s, alpha=alpha, rasterized=True)
                label = "False positives"
                ax.scatter(filtered_i[vals_sort_idx][bad], filtered_j[vals_sort_idx][bad], c="red", s=s, alpha=alpha, rasterized=True, edgecolors='none')#, label=label)
                ax.scatter([-1], [-1], c="red", s=50, label="False positives", rasterized=True, edgecolors='none')
            label = "True positives"
            ax.scatter(filtered_i[vals_sort_idx][first_map_vals == 1], filtered_j[vals_sort_idx][first_map_vals == 1], c=c, s=s, alpha=alpha, rasterized=True, edgecolors='none')#, label=label)
            ax.scatter([-1], [-1], c=c, s=50, label=label, rasterized=True, edgecolors='none')

    ax.set_xlim(0, max_L)
    ax.set_ylim(0, max_L)
    ax.set_aspect('equal')
    ax.set_ylim(ax.get_ylim()[::-1])
    ax.set_xticks([])
    ax.set_yticks([])
    
    return f, ax

def eval_hetero_oligomer(
    cons,
    chain_break_pos,
    L=1,
    monomer_minsep=24,
    cutoffs=[None],
    ss=[5],
    cc=['gray'],
    f=None,
    ax=None,
    vmax=None,
    vmin=0,
    monomer_p_at_k=False
    ):
    # Get complex length
    n = cons[0].shape[0]
    complex_length = cons[0].shape[0]
    #### Get upper triangle matrix for heatmap ####
    upper = np.full((n, n), np.nan)
    for i in range(n):
        for j in range(i, n):  # i <= j
            if j - i >= monomer_minsep:
                upper[i, j] = cons[1][i, j]
            else:
                upper[i, j] = np.nan
    im = ax.imshow(upper, origin="upper", cmap="Greys", rasterized=True, vmin=vmin, vmax=vmax)

    ##### Plot ground truth (n = 0) and predictions #####
    # Monomer
    precision_d = {}
    total_interface_contacts = None
    monomer_top_d = {}
    for i, (con, cutoff, s, c) in enumerate(zip(cons, cutoffs, ss, cc)):
        # Iterate over both chains separately
        for chain_idx in range(2):
            chain_start_idx = chain_idx * chain_break_pos
            if chain_idx == 0:
                chain_end_idx = (chain_idx + 1) * chain_break_pos
            else:
                chain_end_idx = complex_length
            sub_con = deepcopy(con[chain_start_idx:chain_end_idx, chain_start_idx:chain_end_idx])
            triu_idx = np.triu_indices_from(sub_con, 1)
            mask = np.abs(triu_idx[0] - triu_idx[1]) >= monomer_minsep
            filtered_i, filtered_j = triu_idx[0][mask], triu_idx[1][mask]
            vals = sub_con[filtered_i, filtered_j]
            if (i == 0) and monomer_p_at_k:
                monomer_top_d[chain_idx] = vals.sum()
            elif (i == 0) and not monomer_p_at_k:
                monomer_top_d[chain_idx] = sub_con.shape[0] * L
            if cutoff is None:
                top = monomer_top_d[chain_idx]
                if len(vals) > top:  # Make sure we have enough values
                    use_cutoff = np.sort(vals)[::-1][top]
                else:
                    use_cutoff = np.min(vals) if len(vals) > 0 else 0
            else:
                use_cutoff = cutoff
            if np.all((con == 0) | (con == 1)):
                vals_sort_idx = vals == 1
                label = "Ground truth"
            else:
                vals_sort_idx = vals > use_cutoff
                label = "True positives"
            if chain_idx > 0:
                label = None
            if i == 1 and len(cons) > 1:
                # For the second contact map, highlight false positives (where ground truth is 0)
                # We need to match indices between the two contact maps
                if filtered_i.size > 0:  # Make sure we have pairs to check
                    # Get values from first contact map at same positions
                    first_map_vals = cons[0][filtered_i[vals_sort_idx] + chain_start_idx, filtered_j[vals_sort_idx] + chain_start_idx]
                    bad = first_map_vals == 0
                    
                    # Plot false positives
                    ax.scatter(filtered_i[vals_sort_idx][bad] + chain_start_idx, filtered_j[vals_sort_idx][bad] + chain_start_idx, c="red", s=s, alpha=0.8, edgecolors='none', rasterized=True)
                    label = None if chain_idx > 0 else "False positives"
                    ax.scatter([-1], [-1], c="red", s=70, label=label, edgecolors='none', rasterized=True)
                precision_d[f"chain_{chain_idx}"] = (len(bad) - bad.sum()) / len(bad)
                # Plot true positives
                label = None if chain_idx > 0 else "True positives"
                ax.scatter([-1], [-1], c=c, s=70, label=label, edgecolors='none', rasterized=True)
                ax.scatter(filtered_i[vals_sort_idx][first_map_vals == 1] + chain_start_idx, filtered_j[vals_sort_idx][first_map_vals == 1] + chain_start_idx, c=c, s=s, alpha=0.8, edgecolors='none', rasterized=True)
            else:
                # Plot ground truth
                label = None if chain_idx > 0 else "Ground truth"
                ax.scatter([-1], [-1], c=c, s=70, label=label, edgecolors='none', rasterized=True)
                ax.scatter(filtered_i[vals_sort_idx] + chain_start_idx, filtered_j[vals_sort_idx] + chain_start_idx, c=c, s=s, edgecolors='none', rasterized=True)

        # Hetero-oligomer contacts
        sub_con = deepcopy(con[chain_break_pos:, :chain_break_pos])
        if i == 0:
            total_interface_contacts = sub_con.sum()
        if total_interface_contacts == 0:
            continue
        filtered_i, filtered_j = np.indices(sub_con.shape)
        filtered_i, filtered_j = filtered_i.flatten(), filtered_j.flatten()
        vals = sub_con[filtered_i, filtered_j]
        if cutoff is None:
            hetero_top = total_interface_contacts
            if len(vals) > hetero_top:
                use_cutoff = np.sort(vals)[::-1][hetero_top]
            else:
                use_cutoff = np.min(vals) if len(vals) > 0 else 0
        else:
            use_cutoff = cutoff
        if np.all((sub_con == 0) | (sub_con == 1)):
            vals_sort_idx = vals == 1
            label = "Ground truth"
        else:
            vals_sort_idx = vals > use_cutoff
            label = "True positives"
        if i == 1 and len(cons) > 1:
            # For the second contact map, highlight false positives (where ground truth is 0)
            # We need to match indices between the two contact maps
            if filtered_i.size > 0:  # Make sure we have pairs to check
                # Get values from first contact map at same positions
                gt_sub_con = deepcopy(cons[0][chain_break_pos:, :chain_break_pos])
                first_map_vals = gt_sub_con[filtered_i[vals_sort_idx], filtered_j[vals_sort_idx]]
                bad = first_map_vals == 0
                
                # Plot false positives
                ax.scatter(filtered_j[vals_sort_idx][bad], filtered_i[vals_sort_idx][bad] + chain_break_pos, c="red", s=s, alpha=0.8, edgecolors='none', rasterized=True)
            precision_d[f"interface"] = (len(bad) - bad.sum()) / len(bad)
            # Plot true positives
            ax.scatter(filtered_j[vals_sort_idx][first_map_vals == 1], filtered_i[vals_sort_idx][first_map_vals == 1] + chain_break_pos, c=c, s=s, alpha=0.8, edgecolors='none', rasterized=True)
        else:
            # Plot ground truth
            ax.scatter(filtered_j[vals_sort_idx], filtered_i[vals_sort_idx] + chain_break_pos, c=c, s=s, edgecolors='none', rasterized=True)
    ax.set_xlim(0, cons[0].shape[0])
    ax.set_ylim(0, cons[0].shape[0])
    ax.set_aspect('equal')
    ax.set_ylim(ax.get_ylim()[::-1])
    ax.set_xticks([])
    ax.set_yticks([])
    return f, ax, precision_d


import os
import random
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from pathlib import Path
# from shortuuid import uuid

import torch
from torch.nn import CrossEntropyLoss
from torch.utils.data import DataLoader, random_split, Dataset
from torch.optim import Adam, Optimizer
from torch.optim.lr_scheduler import LambdaLR, LRScheduler

import sys
# sys.path.append("../funcs/")
from .dataset import MSADataset, MSAConfindContactMapDataset, CollateAFBatch, CollateMSAConfindContactMapBatch, MSA, aa2tok_d, tok2aa_d, nTokenTypes, trRosettaContactMSADataset, CollatetrRosettaContactMSABatch

def init_dataloaders(
    msa_file_paths,
    max_seq_length,
    max_msa_depth,
    min_msa_depth,
    max_tokens,
    max_seq_length_val_test,
    max_msa_depth_val_test,
    max_tokens_val_test,
    min_msa_depth_val_test,
    batch_size,
    batch_size_val_test,
    num_workers,
    train_prop = 0.8,
    val_size = 100,
    pin_memory = True,
    query_only = False,
    random_query = False,
    min_query_coverage = 0.8,
    n_append_random = 0,
    n_append_random_val_test = 0,
    shifted_random = False,
    data_random_seed = 42,
    noising_rate = 0.15,
    mask_ratio = 0.8,
    mutate_ratio = 0.1,
    keep_ratio = 0.1,
    mutate_pssm = False,
    persistent_workers=True,
    prefetch_factor=2,
    scramble_seq_perc_train = None,
    scramble_col_perc_train = None,
    scramble_seq_perc_val_test = None,
    scramble_col_perc_val_test = None
):
    # Set random seed
    torch.manual_seed(data_random_seed)
    np.random.seed(data_random_seed)
    # Randomly subset sequences for train, val, and test
    nTrain = int(train_prop * len(msa_file_paths))
    nVal = val_size
    nTest = len(msa_file_paths) - nTrain - nVal
    remaining_indices = np.arange(len(msa_file_paths))
    train_seq_indices = np.random.choice(remaining_indices, size=nTrain, replace=False)
    remaining_indices = np.setdiff1d(remaining_indices, train_seq_indices)
    val_seq_indices = np.random.choice(remaining_indices, size=nVal, replace=False)
    remaining_indices = np.setdiff1d(remaining_indices, val_seq_indices)
    test_seq_indices = np.random.choice(remaining_indices, size=nTest, replace=False)
    # Initialize datasets
    train_msa_dataset = MSADataset(
        msa_dir = None,
        msa_paths = msa_file_paths[train_seq_indices],
        max_seq_length = max_seq_length,
        max_msa_depth = max_msa_depth,
        max_tokens = max_tokens,
        random_query = random_query,
        min_query_coverage = min_query_coverage,
        n_append_random = n_append_random,
        shifted_random = shifted_random,
        scramble_seq_perc = scramble_seq_perc_train,
        scramble_col_perc = scramble_col_perc_train
    )
    val_msa_dataset = MSADataset(
        msa_dir = None,
        msa_paths = msa_file_paths[val_seq_indices],
        max_seq_length = max_seq_length_val_test,
        max_msa_depth = max_msa_depth_val_test,
        max_tokens = max_tokens_val_test,
        random_query = random_query,
        min_query_coverage = min_query_coverage,
        n_append_random = n_append_random_val_test,
        shifted_random = shifted_random,
        scramble_seq_perc = scramble_seq_perc_val_test,
        scramble_col_perc = scramble_col_perc_val_test
    )
    test_msa_dataset = MSADataset(
        msa_dir = None,
        msa_paths = msa_file_paths[test_seq_indices],
        max_seq_length = max_seq_length_val_test,
        max_msa_depth = max_msa_depth_val_test,
        max_tokens = max_tokens_val_test,
        random_query = random_query,
        min_query_coverage = min_query_coverage,
        n_append_random = n_append_random_val_test,
        shifted_random = shifted_random,
        scramble_seq_perc = scramble_seq_perc_val_test,
        scramble_col_perc = scramble_col_perc_val_test
    )

    # # Adjust maximum sequence lengths for validation and test datasets
    collate_fn_train = CollateAFBatch(
        max_seq_depth = max_msa_depth + n_append_random,
        max_seq_length = max_seq_length,
        min_seq_depth = min_msa_depth,
        query_only = query_only,
        mask_prob = noising_rate,
        mask_ratio = mask_ratio,
        mutate_ratio = mutate_ratio,
        keep_ratio = keep_ratio,
        mutate_pssm = mutate_pssm,
        n_append_random = n_append_random
    )
    collate_fn_val_test = CollateAFBatch(
        max_seq_depth = max_msa_depth_val_test + n_append_random_val_test,
        max_seq_length = max_seq_length_val_test,
        min_seq_depth = min_msa_depth_val_test,
        query_only = query_only,
        mask_prob = noising_rate,
        mask_ratio = mask_ratio,
        mutate_ratio = mutate_ratio,
        keep_ratio = keep_ratio,
        mutate_pssm = mutate_pssm,
        n_append_random = n_append_random_val_test
    )
    
    # Initialize dataloaders
    msa_dataloader_train = torch.utils.data.DataLoader(
        train_msa_dataset, batch_size=batch_size, shuffle=False, num_workers=num_workers, collate_fn=collate_fn_train, pin_memory=pin_memory, 
        persistent_workers=persistent_workers,
        prefetch_factor=prefetch_factor
    )
    msa_dataloader_val = torch.utils.data.DataLoader(
        val_msa_dataset, batch_size=batch_size_val_test, shuffle=False, num_workers=num_workers, collate_fn=collate_fn_val_test, pin_memory=pin_memory,
        persistent_workers=persistent_workers,
        prefetch_factor=prefetch_factor
    )
    msa_dataloader_test = torch.utils.data.DataLoader(
      test_msa_dataset, batch_size=batch_size_val_test, shuffle=False, num_workers=num_workers, collate_fn=collate_fn_val_test, pin_memory=pin_memory,
      persistent_workers=persistent_workers,
      prefetch_factor=prefetch_factor
    )
    return msa_dataloader_train, msa_dataloader_val, msa_dataloader_test

def init_trRosetta_contact_dataloaders(
    train_paired_paths_l,
    val_paired_paths_l,
    max_seq_length_train,
    max_seq_length_val_test,
    max_msa_depth_train,
    max_msa_depth_val_test,
    min_msa_depth_train,
    min_msa_depth_val_test,
    max_tokens_train,
    max_tokens_val_test,
    batch_size_train,
    batch_size_val_test,
    num_workers,
    pin_memory = True,
    prefetch_factor = 2,
    persistent_workers = True,
    data_random_seed = 42
):
    # Set random seed
    torch.manual_seed(data_random_seed)
    np.random.seed(data_random_seed)
    random.seed(data_random_seed)
    # Initialize datasets
    train_dataset = trRosettaContactMSADataset(
        paired_paths_l = train_paired_paths_l,
        max_seq_length = max_seq_length_train,
        max_msa_depth = max_msa_depth_train,
        min_msa_depth = min_msa_depth_train,
        max_tokens = max_tokens_train,
        random_query = False,
        n_append_random = 0,
        shifted_random = False
    )
    val_dataset = trRosettaContactMSADataset(
        paired_paths_l = val_paired_paths_l,
        max_seq_length = max_seq_length_val_test,
        max_msa_depth = max_msa_depth_val_test,
        min_msa_depth = min_msa_depth_val_test,
        max_tokens = max_tokens_val_test,
        random_query = False,
        n_append_random = 0,
        shifted_random = False
    )

    # Adjust maximum sequence lengths for validation and test datasets
    collate_fn_train = CollatetrRosettaContactMSABatch(
        max_seq_depth = max_msa_depth_train,
        max_seq_length = max_seq_length_train,
        min_seq_depth = min_msa_depth_train,
        pad_tok = aa2tok_d['PAD']
    )
    collate_fn_val = CollatetrRosettaContactMSABatch(
        max_seq_depth = max_msa_depth_val_test,
        max_seq_length = max_seq_length_val_test,
        min_seq_depth = min_msa_depth_val_test,
        pad_tok = aa2tok_d['PAD']
    )

    # Initialize dataloaders
    dataloader_train = torch.utils.data.DataLoader(
        train_dataset, batch_size=batch_size_train, shuffle=True, num_workers=num_workers, collate_fn=collate_fn_train, pin_memory=pin_memory,
        persistent_workers=persistent_workers,
        prefetch_factor=prefetch_factor
    )
    dataloader_val = torch.utils.data.DataLoader(
        val_dataset, batch_size=batch_size_val_test, shuffle=False, num_workers=num_workers, collate_fn=collate_fn_val, pin_memory=pin_memory,
        persistent_workers=persistent_workers,
        prefetch_factor=prefetch_factor
    )
    return dataloader_train, dataloader_val

def evaluate_prediction(logits, batch, device, criterion, query_only = False, loss_weights = None, mean_reduction = False):
    dim_logits = logits.shape[-1]
    pred_logits = logits.view(-1, dim_logits)[batch['masked_idx']]
    if query_only:
        mlm_tgt = batch['msas'][:, 0, :].flatten()[batch['masked_idx']]
    else:
        mlm_tgt = batch['msas'].view(-1)[batch['masked_idx']]
    mlm_loss = criterion(pred_logits.float(), mlm_tgt)
    if loss_weights is not None:
        assert loss_weights.shape == mlm_loss.shape
        mlm_loss *= loss_weights
        if mean_reduction:
            mlm_loss = mlm_loss.mean()
    perplexity = torch.exp(mlm_loss)
    accuracy = (pred_logits.argmax(dim = -1) == mlm_tgt).float().mean()
    loss_d = {"mlm_loss": mlm_loss, "accuracy": accuracy, "perplexity": perplexity}
    return loss_d

# Tracking moving averages
class LossTracker:
    def __init__(self, categories_l, ema_alpha = 0.05):
        self.categories_l = categories_l
        self.ema_alpha = ema_alpha
        self.ema_d = {}
        for category in self.categories_l:
            self.ema_d[category] = None

    def update_single(self, category_name, value):
        if category_name not in self.categories_l:
            raise ValueError(f"Category {category_name} not found in categories list")
        value = float(value)
        if self.ema_d[category_name] is None:
            self.ema_d[category_name] = value
        else:
            self.ema_d[category_name] = self.ema_alpha * value + (1 - self.ema_alpha) * self.ema_d[category_name]

    def get_single(self, category_name):
        if category_name not in self.categories_l:
            raise ValueError(f"Category {category_name} not found in categories list")
        if self.ema_d[category_name] is None:
            raise ValueError(f"No values recorded yet for category {category_name}")
        return self.ema_d[category_name]

    def update_multiple(self, loss_d):
        for category_name, value in loss_d.items():
            if category_name not in self.categories_l:
                continue  # Skip unknown categories silently
            self.update_single(category_name, value)

    def get_all(self):
        # Filter out None values
        return {k: v for k, v in self.ema_d.items() if v is not None}

    def reset(self):
        for category_name in self.categories_l:
            self.ema_d[category_name] = None

# Tracks average values over gradient accumulation steps
class GradAccumStatTracker:
    def __init__(self, stats_l):
        self.stats_l = stats_l
        self.nSteps = 0
        self.stat_d = {}
        for stat in self.stats_l:
            self.stat_d[stat] = 0

    def increment_step(self):
        self.nSteps += 1

    def update_single(self, stat_name, value):
        value = float(value)
        self.stat_d[stat_name] += value

    def add_category(self, stat_name):
        self.stats_l.append(stat_name)
        self.stat_d[stat_name] = 0

    def get_single(self, stat_name):
        return self.stat_d[stat_name] / self.nSteps

    def update_multiple(self, update_stat_d):
        for stat_name, value in update_stat_d.items():
            self.update_single(stat_name, value)

    def get_all(self):
        return {k: v / self.nSteps for k, v in self.stat_d.items()}

    def reset(self):
        for stat_name in self.stats_l:
            self.stat_d[stat_name] = 0
        self.nSteps = 0


class GradAccumLossTracker:
    def __init__(self):
        self.total_loss = 0
        self.total_tokens = 0
        self.total_correct = 0
        self.nSteps = 0
    
    def update_total_loss(self, loss):
        self.total_loss += loss
    
    def update_total_tokens(self, tokens):
        self.total_tokens += tokens

    def update_total_correct(self, correct):
        self.total_correct += correct
    
    def get_final_loss(self):
        return self.total_loss / self.total_tokens
    
    def get_final_accuracy(self):
        return self.total_correct / self.total_tokens

    def get_final_perplexity(self):
        with torch.no_grad():
            return torch.exp(self.get_final_loss())

    def increment_step(self):
        self.nSteps += 1

    def reset(self):
        self.total_loss = 0
        self.total_tokens = 0
        self.total_correct = 0
        self.nSteps = 0

def set_seed(seed):
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    np.random.seed(seed)
    random.seed(seed)



import torch
import torch.nn.functional as F
from torch.nn import Module
from typing import List
import einx
from einops import pack
from .core import LinearNoBias
from .custom_typing import (
    Float
)

# Molecule features
"""
molecule_feats: [*, 5]:
0: molecule_index
1: token_index
2: asym_id
3: entity_id
4: sym_id
"""
MOLECULE_FEATS = 5

class RelativePositionEncoding(Module):
    """
    Clipped relative positional encoding to initialize the pair representation
    Some of this is actually a bit unnecessary, but was initially framed to support multi-chain MSAs
    The current MSA Pairformer release does not actually support multi-chain MSAs, but this is left here for future use
    """
    def __init__(
        self,
        r_max = 32,
        s_max = 2,
        dim_out = 128
    ):
        super().__init__()
        self.r_max = r_max
        self.s_max = s_max

        dim_input = (2*r_max + 2) + (2*r_max + 2) + 1 + (2*s_max + 2)
        self.out_embedder = LinearNoBias(dim_input, dim_out)

    def forward(
        self,
        batch_size: int, 
        seq_len: int,
        device: torch.device,
        complex_chain_break_indices: List[List[int]] | None = None
    ) -> Float['b n n dp']:

        # One-hot encode distances
        # bins will be single-offset distances
        def onehot(x, bins):
            dist_from_bins = einx.subtract('... i, j -> ... i j', x, bins)
            indices = dist_from_bins.abs().min(dim = -1, keepdim=True).indices
            one_hots = F.one_hot(indices.long(), num_classes = len(bins))
            return one_hots

        # Initialize token indices
        token_idx = torch.arange(seq_len).unsqueeze(0).expand(batch_size, seq_len)
        # complex_chain_break_indices is a list of lists
        # Each list contains chain break indices for a given complex (there may be multiple chain breaks per complex)
        if complex_chain_break_indices is not None:
            for batch_idx, chain_break_indices in enumerate(complex_chain_break_indices):
                if len(chain_break_indices) > 0:
                    for chain_break_idx in chain_break_indices:
                        token_idx[batch_idx, chain_break_idx:] += 100
        # Compute residue, token, and chain distances
        diff_token_idx = einx.subtract('b i, b j -> b i j', token_idx, token_idx)
        # Mask for same residue, chain, and entity
        mask_same_entity = torch.ones((batch_size, seq_len, seq_len, 1)).to(device)
        # Compute clipped distances
        d_token = torch.clip(diff_token_idx + self.r_max, 0, 2 * self.r_max).to(device)
        d_res = torch.full((batch_size, seq_len, seq_len), self.r_max, device=device)
        d_chain = torch.full((batch_size, seq_len, seq_len), 2*self.s_max+1, device=device)
        # Define bins
        r_arange = torch.arange(2*self.r_max + 2, device=device)
        s_arange = torch.arange(2*self.s_max + 2, device=device)
        # Assign 1-hot encoding of distances
        a_rel_pos = onehot(d_res, r_arange)
        a_rel_token = onehot(d_token, r_arange)
        a_rel_chain = onehot(d_chain, s_arange)
        # Concatenate tensors and project
        out, _ = pack((a_rel_pos, a_rel_token, mask_same_entity, a_rel_chain), 'b i j *')
        
        return self.out_embedder(out)


import torch
import torch.nn.functional as F
from torch.nn import Module, LayerNorm, Linear, Parameter, Sigmoid, ModuleList
from einops import rearrange
from .core import SwiGLU
from .custom_typing import (
    Float,
    Bool,
    typecheck
)

##########################
# Language modeling head #
##########################
class LMHead(Module):
    def __init__(
        self,
        dim_msa,
        dim_output,
    ):
        super().__init__()
        self.init_ln = LayerNorm(dim_msa)
        self.dense = Linear(dim_msa, dim_msa*2)
        self.dense_activation = SwiGLU()
        self.pre_logit_norm = LayerNorm(dim_msa)
        self.bias = Parameter(torch.zeros(dim_output))
        self.weight = Linear(dim_msa, dim_output, bias=False).weight

    def forward(
        self,
        msa_repr: Float["b s n d"]
    ) -> Float["b s n *"]:
        x = self.init_ln(msa_repr)
        x = self.dense(x)
        x = self.dense_activation(x)
        x = self.pre_logit_norm(x)
        x = F.linear(x, self.weight) + self.bias
        return x

################
# Contact head #
################
class LogisticRegressionContactHead(Module):
    def __init__(
        self,
        dim_pairwise,
    ):
        super().__init__()
        self.init_ln = LayerNorm(dim_pairwise)
        self.bias = Parameter(torch.zeros(1))
        self.weight = Linear(dim_pairwise, 1, bias=False).weight
        self.sigmoid = Sigmoid()

    def forward(
        self,
        pair_repr: Float["b n n d"]
    ) -> Float["b s n *"]:
        x = self.init_ln(pair_repr)
        x = F.linear(x, self.weight) + self.bias
        x = self.sigmoid(x)
        # Symmetrize the output matrix
        x = x.squeeze(-1)
        x = (0.5 * (x + x.transpose(-1, -2)))
        return x